local version = require "version"
local socket = require "socket"
local system = require "love.system"
local timer = require "love.timer"
local fs = require "love.filesystem"
local time = timer.getTime
local mime = require "mime"
local json = require "lunajson"
local channel = require "channel"
local deckLib = require "deckLib"
local platform = system.getOS()
local ceil = math.ceil
local floor = math.floor
local format = string.format
local cfg = ...
local server
local tcpClient = {} -- connected decks
local uiVars = {} -- web UI vars
local appVars = {}
local schJobs
local mime = {
	html = "text/html",
	css = "text/css",
	png = "image/png",
}
local serverId = "EvilPenguin"
require "fileLib"


-- used by pcall
local function jsonDecode(data)
	return json.decode(data, nil, "___null___")
end


local function debug(data)
	channel.main:push(json.encode({ id = "debug", data = data }))
end


local function reset()
	if server and server.close then server:close() end
	server = nil
end


local function saveUIVar(ip, var, val)
	if not uiVars[ip] then uiVars[ip] = {} end
	uiVars[ip][var] = val
end


local function uiVar(ip, var)
	return uiVars[ip] and uiVars[ip][var] or nil
end


local function compileCmd(data)
	data = (data or "").."\n\n"
	local desc, cmd
	local result = ""
	while true do
		data = data:gsub("^[\n]+", "")
		desc, cmd = data:match("^([^\n]-)\n(.-)\n\n")
		if not desc or not cmd then break end
		cmd = cmd:gsub("\n", "\t")
		if result ~= "" then result = result.."\n" end
		result = result..cmd
		data = data:gsub("^([^\n]-)\n(.-)\n\n", "")
	end
	return result
end


local function startSchedule(schedule)
	if not schedule then return nil, "Error - try reselecting schedule" end
	schedule = schedule:gsub("%W", "_")
	local schData = fs.read("Scheduler/"..schedule)
	if not schData then return nil, "Error - could not read schedule" end
	local check
	check, schData = pcall(jsonDecode, schData)
	if not check then return nil, "Error - could not decode schedule" end
	-- copy schedule
	schJobs = schData
	-- set target times for each job
	local time = os.time()
	for k, v in pairs(schJobs) do
		if v.live then
			v.target = time + (tonumber(v.delay) or 0)
			v.cmd = compileCmd(v.cmd)
			if cfg.debug then debug(format("Scheduled: %s", k)) end
		end
	end
	appVars.actSchedule = schedule
	appVars.actScheduleDispName = schedule:gsub("_", " ")
	return true
end


local function stopSchedule()
	schJobs = nil
	appVars.actSchedule = nil
	appVars.actScheduleDispName = nil
end


-- process post data from web UI
local function process(ip, post)
	if not post.id then
		return nil
	elseif post.id == "cfg" and post.action then
		-- cfg / setup screen
		local alert = ""
		if post.action == "save" then
			local xPort, xPass, xTout, xWebport, xWebpass, xWebopen = cfg.port, cfg.pass, cfg.tout, cfg.webport, cfg.webpass, cfg.webopen
			cfg = {
				port = tonumber(post.port) or 4445,
				pass = post.pass or "",
				tout = tonumber(post.tout) or 0,
				sfx = post.sfx and true or false,
				debug = post.debug and true or false,
				webport = tonumber(post.webport) or 10314,
				webpass = post.webpass or "",
				webopen = post.webopen and true or false,
			}
			if cfg.port < 1024 then cfg.port = 4445 end
			if cfg.tout < 10 then cfg.tout = 0 end
			alert = alert.."toastr['success']('Settings applied successfully')\n"

			-- sfx status and cfg to tcp thread
			channel.tcp:push(json.encode({ id = "buff", data = "obsfx\t"..(cfg.sfx and 0 or 1).."\n" }))
			channel.tcp:push(json.encode({ id = "cfg", data = cfg }))
			channel.main:push(json.encode({ id = "cfg", data = cfg }))

			local success, msg = fs.write("updeck.txt", json.encode(cfg))
			if success then
				if cfg.debug then debug("Config saved OK") end
				alert = alert.."toastr['success']('Settings saved to file successfully')\n"
			else
				if cfg.debug then debug("Failed to save config", msg) end
				alert = alert.."toastr['error']('Settings saved to file successfully')\n"
			end
			if cfg.port ~= xPort or cfg.pass ~= xPass or cfg.tout ~= xTout or xWebport ~= cfg.webport or xWebopen ~= cfg.webopen then
				channel.main:push(json.encode({ id = "reset" }))
				channel.http:push(json.encode({ id = "reset" }))
				alert = alert.."toastr['warning']('Connection settings changed - resetting UPDeck')\n"
			end
			love.system.setClipboardText(fs.getSaveDirectory())
		elseif post.action == "reset" then
			channel.main:push(json.encode({ id = "reset" }))
			alert = alert.."toastr['warning']('Resetting UPDeck')\n"
		elseif post.action == "files" then
			channel.main:push(json.encode({ id = "files" }))
			alert = alert.."toastr['success']('Opening UPDeck folder')\n"
		end
		return alert
	elseif post.id == "deckAdd" and post.name then
		local newDeck = post.name:gsub("%W", "_")
		if fs.getInfo("Backup/"..newDeck) then
			return "toastr['error']('Deck already exists')\n"
		elseif fs.write("Backup/"..newDeck, json.encode({ buttonData = {}, groupData = {}, sliderData = {} })) then
			return "toastr['success']('Deck created')\npostForm({ id : 'util', action : 'decks' })\n";
		else
			return "toastr['error']('Error creating deck')\n"
		end
	elseif post.id == "deck" then
		-- select deck
		saveUIVar(ip, "deck", post.deckSelect)
		return
			"$('#webDeck').load('/webDeck', fun".."ction(){ $('#webDeck').resize() })\n"..
			"$('#volCfg').load('/volCfg', fun".."ction(){ $('#volCfg').resize() })\n"
	elseif post.id == "deckOpen" then
		local xdeck = uiVars[ip].deck
		if xdeck then
			return format(
				"window.open('/live/%s/%d/%d/%d/%d', '%s', 'width=600, height=550', true)\n",
				xdeck,
				uiVars[ip]["liveGridW_"..xdeck] or 5,
				uiVars[ip]["liveGridH_"..xdeck] or 4,
				uiVars[ip]["liveBank_"..xdeck] or 1,
				uiVars[ip]["liveIconSz_"..xdeck] or 64,
				xdeck
			)
		else
			return "toastr['error']('Select a deck')\n"
		end
	elseif post.id == "deckExec" then
		saveUIVar(ip, "liveButton", tonumber(post.button))
		local xdeck, button = post.deck or "", tostring(uiVars[ip].liveButton)
		if not xdeck or not button then return "toastr['error']('Error - reselect deck')\n" end
		if cfg.debug then debug(format("deckExec : %s button #%d", xdeck, button)) end
		local cmd, msg, tgLabel, tgIconData = deckLib.press(xdeck, button)
		if not cmd then return "toastr['error']("..msg..")\n" end
		if cmd ~= "" then channel.tcp:push(json.encode({ id = "xdeck", deck = xdeck, data = cmd.."\n" })) end
		local resp = ""
		if tgLabel then resp = resp..format("$('#label%d').html('%s')\n", button, tgLabel) end
		if tgIconData then resp = resp..format("$('#icon%d').attr('src', '%s')\n", button, tgIconData) end
		if resp ~= "" then return resp end
	elseif post.id == "deckBank" then 
		-- select bank
		saveUIVar(ip, "bank", post.bank)
		if uiVars[ip].deck and uiVars[ip].bank then
			return "$('#webDeck').load('/webDeck', fun".."ction(){ $('#webDeck').resize() })\n"
		end
	elseif post.id == "liveUpdate" then 
		local xdeck = post.deck
		if xdeck then
			if post.bank then saveUIVar(ip, "liveBank_"..xdeck, tonumber(post.bank) or 1) end
			if post.gridW then saveUIVar(ip, "liveGridW_"..xdeck, tonumber(post.gridW) or 5) end
			if post.gridH then saveUIVar(ip, "liveGridH_"..xdeck, tonumber(post.gridH) or 4) end
			if post.iconSz then saveUIVar(ip, "liveIconSz_"..xdeck, tonumber(post.iconSz) or 64) end
			if post.controls then saveUIVar(ip, "liveControls_"..xdeck, tonumber(post.controls) or 0) end
			return format("$('#liveDeck').load('%s', function(){ $('#liveDeck').resize() })\n", "/liveDeck/"..xdeck)
		end
	elseif post.id == "deckBtn" then
		-- select button
		local button = post.button or 1
		saveUIVar(ip, "button", button)
		local check
		local deck = fs.read("Backup/"..(uiVars[ip].deck or ""))
		uiVars[ip].iconData = {}
		if deck then check, deck = pcall(jsonDecode, deck) end
		if check and deck.buttonData then
			uiVars[ip].buttonData = { deck.buttonData[button] or {}, deck.buttonData[format("%s", tonumber(button) + 1000)] or {} }
			local icon
			if deck.iconData then
				icon = uiVars[ip].buttonData[1].icon
				if icon and deck.iconData[icon] then uiVars[ip].iconData[1] = deck.iconData[icon] end
				icon = uiVars[ip].buttonData[2].icon
				if icon and deck.iconData[icon] then uiVars[ip].iconData[2] = deck.iconData[icon] end
			end
		else
			uiVars[ip].buttonData = { {}, {} }
		end
	elseif post.id == "editBtn" and uiVars[ip] then
		local buttonData = uiVars[ip].buttonData
		local iconData = uiVars[ip].iconData
		if post.action == "fetch" and buttonData then
			-- fetch button data (both button states)
			local hilite1 = tonumber(buttonData[1].hilite) or 1
			local hilite2 = tonumber(buttonData[2].hilite) or 1
			local resp = 
				format("$('#label1').val('%s')\n", (buttonData[1].label or ""):gsub("\\", "\\\\"):gsub("\n", " "):gsub("'", "\\\'"))..
				format("$('#label2').val('%s')\n", (buttonData[2].label or ""):gsub("\\", "\\\\"):gsub("\n", " "):gsub("'", "\\\'"))..
				format("$('#group1').val('%s')\n", (buttonData[1].group or ""):gsub("\\", "\\\\"):gsub("\n", " "):gsub("'", "\\\'"))..
				format("$('#group2').val('%s')\n", (buttonData[2].group or ""):gsub("\\", "\\\\"):gsub("\n", " "):gsub("'", "\\\'"))..
				format("$('#hilite1').val('%s')\n", hilite1)..
				format("$('#hilite2').val('%s')\n", hilite2)
			if iconData[1] then
				resp = resp..
					format("$('#img1').attr('src', 'data:image/png;base64,%s')\n", iconData[1])..
					"upCrop1.croppie('bind', { url: $('#img1').attr('src') })\n"
			end
			if iconData[2] then
				resp = resp..
					format("$('#img2').attr('src', 'data:image/png;base64,%s')\n", iconData[2])..
					"upCrop2.croppie('bind', { url: $('#img2').attr('src') })\n"
			end
			for i = 1, 8 do
				if hilite1 == i then
					resp = resp..format("$('#hilite1%d').css({opacity:1.0}).addClass('hiliteBorder2').removeClass('hiliteBorder1')\n", i)
				else
					resp = resp..format("$('#hilite1%d').css({opacity:0.4}).addClass('hiliteBorder1').removeClass('hiliteBorder2')\n", i)
				end
				if hilite2 == i then
					resp = resp..format("$('#hilite2%d').css({opacity:1.0}).addClass('hiliteBorder2').removeClass('hiliteBorder1')\n", i)
				else
					resp = resp..format("$('#hilite2%d').css({opacity:0.4}).addClass('hiliteBorder1').removeClass('hiliteBorder2')\n", i)
				end
			end

			-- build the commands text
			for j = 1, 2 do
				local cmd = ""
				local cmdcfg = buttonData[j].cfg or {}
				for i, t in ipairs(cmdcfg) do
					cmd = cmd..(t.desc or "No description").."\n"..(t.cmd or ""):gsub("\t", "\n").."\n"
					-- separate each command with a blank line
					while cmd:sub(-2) ~= "\n\n" do cmd = cmd.."\n" end
				end
				-- remove trailing newlines
				while cmd:sub(-1) == "\n" do cmd = cmd:sub(1, -2) end
				-- escape special characters
				cmd = cmd:gsub("\\", "\\\\"):gsub("'", "\\\'"):gsub("\n", "\\n")
				resp = resp..format("$('#cmd%d').val('%s')\n", j, cmd)
			end

			-- normal or toggle
			if (buttonData[1].group or "") == "*" then resp = resp.."$('#toggle').trigger('click')\n" end

			return resp
		elseif post.action == "save" then
			-- allow saving selected button as another button
			local saveButton = tonumber(post.button)
			if saveButton then saveUIVar(ip, "button", saveButton) end
			local button = uiVars[ip].button
			if not button then return "toastr['error']('An error occurred - please reselect a deck button')\n" end

			-- only normal deck buttons can have more than 1 state
			local bIndex = { tostring(button) }
			local _button = tonumber(button)
			if post.toggle and _button and _button > 0 and _button < 1000 then
				bIndex[2] = tostring(_button + 1000)
			end

			local btnData = {}
			local icnData = {}
			for i = 1, #bIndex do
				local targetButton = bIndex[i]
				local data = (post["cmd"..i] or "").."\n\n"
				local desc, cmd
				local btnCmd, btnCfg = "", {}
				local imgData, iconName
				while true do
					-- reformat button commands into command and config strings
					data = data:gsub("^[\n]+", "")
					desc, cmd = data:match("^([^\n]-)\n(.-)\n\n")
					if not desc or not cmd then break end
					cmd = cmd:gsub("\n", "\t")
					if btnCmd ~= "" then btnCmd = btnCmd.."\n" end
					btnCmd = btnCmd..cmd
					btnCfg[#btnCfg + 1] = { desc = desc, cmd = cmd }
					data = data:gsub("^([^\n]-)\n(.-)\n\n", "")
				end
				btnData[targetButton] = {
					label = post["label"..i] or "",
					group = post.toggle and "*" or post["group"..i] or "",
					hilite = tonumber(post["hilite"..i]) or 1,
					cmd = btnCmd,
					cfg = btnCfg,
				}
				
				-- icon update
				imgData = post["imgData"..i] or ""
				if imgData:sub(1, 22) == "data:image/png;base64," then
					-- icon used
					imgData = imgData:sub(23)
				else
					-- blank
					imgData = ""
				end
				iconName = format('%d-%s.png', targetButton, floor(socket.gettime() * 1000))
				icnData[targetButton] = { name=iconName, data=imgData }
			end
			
			-- build response string
			local resp = format("$('#buttonTarget').text('%s')\n", button)

			-- update deck backup file
			local deck = uiVars[ip].deck
			if not deck then
				resp = resp.."toastr['error']('No deck is currently selected - deck file not updated')\n"
			else
				local data = fs.read("Backup/"..deck)
				if not data then
					resp = resp.."toastr['error']('Could not open deck file - not updated')\n"
				else
					local check
					check, data = pcall(jsonDecode, data)
					if check and type(data) == "table" and type(data.buttonData) == "table" and type(data.groupData) == "table" and type(data.sliderData) == "table" then
						if type(data.iconData) ~= "table" then data.iconData = {} end
						for i = 1, #bIndex do
							local targetButton = bIndex[i]

							-- update button data
							if not data.buttonData[targetButton] then data.buttonData[targetButton] = {} end
							for k, v in pairs(btnData[targetButton]) do
								data.buttonData[targetButton][k] = v
							end

							-- delete existing icon
							data.buttonData[targetButton].icon = nil
							local del = {}
							local prefix = targetButton.."-"
							for k2, v2 in pairs(data.iconData) do
								if k2:sub(1, prefix:len()) == prefix then del[#del + 1] = k2 end
							end
							for _, k2 in ipairs(del) do data.iconData[k2] = nil end

							-- update icon
							if icnData[targetButton].data ~= "" then
								data.buttonData[targetButton].icon = icnData[targetButton].name
								data.iconData[icnData[targetButton].name] = icnData[targetButton].data
							end
						end
						-- save to file
						local success, msg = fs.write("Backup/"..deck, json.encode(data, "___null___"))
						if success then
							resp = resp.."toastr['success']('Changes saved to deck file')\n"
						else
							resp = resp.."toastr['error']('Could not write to deck file - not updated')\n"
						end
						-- update device
						uiVars[ip].client = post.client
						local client = uiVars[ip].client
						if not client or client == "" then
							resp = resp.."toastr['warning']('No device has been selected to update')\n"
						elseif not tcpClient[client] then
							resp = resp.."toastr['error']('Selected device is no longer connected')\n"
						else
							channel.tcp:push(json.encode({ id = "btnupdate", data = {client = client, data = {btnData = btnData, icnData = icnData}} }))
							resp = resp..format("toastr['success']('Changes sent to device : %s')\n", tcpClient[client] or "?")
						end
					else
						resp = resp.."toastr['error']('Invalid deck file - not updated')\n"
					end
				end
			end	

			-- return alerts to web UI
			resp = resp.."window.scrollTo(0,0)\n"
			return resp
		end
	elseif post.id == "swap" then
		--[[
		-- swap buttons
		local b1 = tostring(post.b1)
		local b2 = tostring(post.b2)
		if not b1 or not b2 then return "toastr['error']('Reselect 2 buttons to swap')\n" end
		local b3 = tostring((tonumber(b1) or 0) + 1000)
		local b4 = tostring((tonumber(b2) or 0) + 1000)
		local deck = uiVar(ip, "deck")
		if not deck then return "toastr['error']('Try reselecting deck')\n" end
		deck = fs.read("Backup/"..deck)
		if not deck then return "toastr['error']('Could not read deck file - try reselecting deck')\n" end
		local check
		check, deck = pcall(jsonDecode, deck) end
		if not check or type(deck.buttonData) ~= "table" or type(deck.iconData) ~= "table" then
			return "toastr['error']('Error decoding deck file - try reselecting deck')\n"
		end
		local swapData = {
			btnData = {
				[b1] = deck.buttonData[b2] or {},
				[b3] = deck.buttonData[b4] or {},
				[b2] = deck.buttonData[b1] or {},
				[b4] = deck.buttonData[b3] or {},
			},
			icnData = {
			},
		}
		local xicon, icon
		for k, v in pairs(swapData.btnData) do
			xicon = swapData.btnData[k].icon
			if xicon and deck.iconData[xicon] then
				icon = format('%d-%s.png', tonumber(k) or 0, floor(socket.gettime() * 1000))
				swapData.icnData[icon] = deck.iconData[xicon]
				deck.iconData[icon] = deck.iconData[xicon]
				deck.iconData[xicon] = nil
				v.icon = icon
				deck.buttonData[k] = v
			end
		end
		-- save to file
		local success, msg = fs.write("Backup/"..uiVar(ip, "deck"), json.encode(deck, "___null___"))
		if success then
			resp = resp.."toastr['success']('Button swap saved to deck file')\n"
		else
			resp = resp.."toastr['error']('Could not write to deck file - not updated')\n"
		end
		--saveUIVar(ip, "button", b2)
		--saveUIVar(ip, "buttonData", { deck.buttonData[b2] or {}, deck.buttonData[format("%s", tonumber(b2) + 1000)] or {} }
		--]]
	elseif post.id == "scheduler" then
		-- select schedule
		saveUIVar(ip, "schedule", post.schSelect)
		saveUIVar(ip, "scheduleDispName", (post.schSelect or ""):gsub("_", " "))
		return 
			"$('#schAct').load('/schAct', fun".."ction(){ $('#schAct').resize() })\n"..
			"$('#schCfg').load('/schCfg', fun".."ction(){ $('#schCfg').resize() })\n"
	elseif post.id == "schedule" then
		if post.action == "save" then
			-- edit schedule
			local scheduleName = (post.name or ""):gsub("%W", "_")
			local existingName = uiVars[ip].schedule
			local resp = ""
			if scheduleName == "" then
				resp = resp.."toastr['error']('Invalid schedule name')\n"
			elseif not existingName then
				-- create new schedule
				local info = fs.getInfo("Scheduler/"..scheduleName)
				if info then
					-- already exists
					resp = resp.."toastr['error']('A schedule already exists with that name')\n"
				else
					-- create new file
					local success = fs.write("Scheduler/"..scheduleName, json.encode({}, "___null___"))
					if success then
						resp = resp.."toastr['success']('Schedule created')\n"
						saveUIVar(ip, "schedule", scheduleName)
						saveUIVar(ip, "scheduleDispName", scheduleName:gsub("_", " "))
					else
						resp = resp.."toastr['error']('Error creating schedule')\n"
					end
				end
			elseif existingName ~= scheduleName then
				-- change name
				local info = fs.getInfo("Scheduler/"..scheduleName)
				if info then
					-- already exists
					resp = resp.."toastr['error']('A schedule already exists with that name')\n"
				else
					local schData = fs.read("Scheduler/"..existingName)
					if not schData then
						resp = resp.."toastr['error']('Error reading existing schedule')\n"
					else
						local success = fs.write("Scheduler/"..scheduleName, schData)
						if success then
							resp = resp.."toastr['success']('Schedule renamed')\n"
							success = fs.remove("Scheduler/"..existingName)
							if not success then
								-- old file still exists
								resp = resp.."toastr['warning']('Error removing file with old name')\n"
							end
							saveUIVar(ip, "schedule", scheduleName)
							saveUIVar(ip, "scheduleDispName", scheduleName:gsub("_", " "))
							if existingName == appVars.actSchedule then
								appVars.actSchedule = scheduleName
								appVars.actScheduleDispName = uiVar(ip, "scheduleDispName")
							end
						else
							resp = resp.."toastr['error']('Error renaming schedule')\n"
						end
					end
				end
			else
				-- no change
				resp = resp.."toastr['warning']('No changes made')\n"
			end
			resp = resp.."postForm({ id : 'util', action : 'schedules' }, fun".."ction(data){ eval(data); postForm('scheduler') } )\n"
			return resp
		elseif post.action == "delete" then
			local scheduleName = uiVar(ip, "schedule")
			if not scheduleName then
				return "toastr['error']('No schedule selected to delete - try reselecting a schedule')\n"
			end
			local success = fs.remove("Scheduler/"..scheduleName)
			if not success then
				-- old file still exists
				return "toastr['error']('Could not delete schedule')\n"
			else
				saveUIVar(ip, "schedule", nil)
				return
					"toastr['success']('Schedule deleted')\n"..
					"postForm({ id : 'util', action : 'schedules' }, fun".."ction(data){ eval(data); $('#new').click() } )\n"
			end
		elseif post.action == "activate" then
			local success, msg = startSchedule(uiVar(ip, "schedule"))
			if success then
				return
					"$('#schAct').load('/schAct', fun".."ction(){ $('#schAct').resize() })\n"..
					"toastr['success']('Schedule activated')\n"
			end
			return format("toastr['error']('%s')\n", msg)
		elseif post.action == "stop" then
			if appVars.actSchedule then
				stopSchedule()
				return
					"$('#schAct').load('/schAct', fun".."ction(){ $('#schAct').resize() })\n"..
					"toastr['success']('Schedule stopped')\n"
			else
				return "toastr['error']('Error - no schedule to stop')\n"
			end
		end
	elseif post.id == "job" then
		if post.action == "select" then
			-- select job to edit
			saveUIVar(ip, "job", post.job)
			saveUIVar(ip, "jobDispName", (post.job or ""):gsub("_", " "))
			saveUIVar(ip, "jobName", post.job and uiVar(ip, "jobDispName") or "New job")
			return "window.location.href='/job'\n"
		elseif post.action == "fetch" then
			-- fetch job data
			local schedule = uiVar(ip, "schedule")
			local job = uiVar(ip, "job") or ""
			local jobDispName = uiVar(ip, "jobDispName") or "New job"
			if not schedule then return "toastr['error']('Error - try reselecting scheduler / job')\n" end
			resp = "$('#name').val('"..jobDispName.."')\n"
			local schData, check
			schData = fs.read("Scheduler/"..schedule)
			if not schData then return resp.."toastr['error']('Error - could not read schedule file')\n" end
			check, schData = pcall(jsonDecode, schData)
			if not check or type(schData) ~= "table" then return resp.."toastr['error']('Error - could not decode schedule file')\n" end
			local jobData = schData[job]
			if type(jobData) ~= "table" then
				if uiVar(ip, "job") then resp = resp.."toastr['error']('Error - no job data found')\n" end
				jobData = { interval = 300, delay = 0, live = true, cmd = "" }
			end
			return
				resp..
				"$('#jobName').text('"..jobDispName.."')\n"..
				"$('#interval').val("..(tonumber(jobData.interval) or 300)..")\n"..
				"$('#delay').val("..(tonumber(jobData.delay) or 0)..")\n"..
				--"$('#count').val("..(tonumber(jobData.count) or 0)..")\n"..
				--"$('#max').val("..(tonumber(jobData.max) or 0)..")\n"..
				"$('#live').prop('checked', "..(jobData.live and 'true' or 'false')..")\n"..
				"$('#cmd').val('"..(tostring(jobData.cmd) or ""):gsub("\\", "\\\\"):gsub("\n", "\\n"):gsub("'", "\\\'").."')\n"
		elseif post.action == "save" then
			local schedule = uiVar(ip, "schedule")
			local isLive = schedule == appVars.actSchedule
			local job = uiVar(ip, "job")
			if not schedule then return "toastr['error']('Error - try reselecting schedule')\n" end
			local schData, check
			schData = fs.read("Scheduler/"..schedule)
			if not schData then return "toastr['error']('Error - try reselecting schedule')\n" end
			check, schData = pcall(jsonDecode, schData)
			if not check or type(schData) ~= "table" then return "toastr['error']('Error - could not decode schedule file')\n" end
			local jobName = (post.name or ""):gsub("%W", "_")
			if jobName == "" then return "toastr['error']('Invalid job name')\n" end
			if jobName ~= job then
				if schData[jobName] then
					return "toastr['error']('A job already exists with that name')\n"
				else
					if job then
						schData[job] = nil
						if isLive then schJobs[job] = nil end
					end
				end
			end
			local interval = tonumber(post.interval) or 300
			if interval < 1 then interval = 1 end
			local delay = tonumber(post.delay) or 0
			if delay < 0 then delay = 0 end
			schData[jobName] = { interval = interval, delay = delay, cmd = tostring(post.cmd) or "" }
			if post.live then schData[jobName].live = true end
			if isLive then
				-- update active schedule
				schJobs[jobName] = {
					interval = interval,
					delay = delay,
					live = post.live and true or false,
					cmd = compileCmd(post.cmd),
					target = os.time() + delay
				}
			end
			local resp = ""
			-- save to file
			local success, msg = fs.write("Scheduler/"..schedule, json.encode(schData, "___null___"))
			if success then
				resp = resp.."toastr['success']('Changes saved to schedule')\npostForm({ id : 'job', action : 'fetch' })\n"
				saveUIVar(ip, "job", jobName)
				saveUIVar(ip, "jobDispName", jobName:gsub("_", " "))
			else
				resp = resp.."toastr['error']('Could not write to schedule - not updated')\n"
			end
			-- return alerts to web UI
			resp = resp.."window.scrollTo(0,0)\n"
			return resp
		elseif post.action == "delete" then
			local schedule = uiVar(ip, "schedule")
			local job = uiVar(ip, "job")
			if not schedule or not job then return "toastr['error']('Error - try reselecting schedule / job')\n" end
			local schData, check
			schData = fs.read("Scheduler/"..schedule)
			if not schData then return "toastr['error']('Error - try reselecting schedule')\n" end
			check, schData = pcall(jsonDecode, schData)
			if not check or type(schData) ~= "table" then return "toastr['error']('Error - could not decode schedule file')\n" end
			-- delete job and save to file
			schData[job] = nil
			local success, msg = fs.write("Scheduler/"..schedule, json.encode(schData, "___null___"))
			if success then
				saveUIVar(ip, "job", nil)
				saveUIVar(ip, "jobDispName", nil)
				return "toastr['success']('Job deleted from schedule')\npostForm({ id : 'job', action : 'fetch' })\n"
			else
				return "toastr['error']('Could not write to schedule - not deleted')\n"
			end
		end
	elseif post.id == "combob" then
		if post.action == "export" then
			local combob = require "combobulator"
			local success, result = combob.export(post.data, post.name, post.deck, post.schedule, post.sfx)
			if not success then
				return "toastr['error']('An error occurred - see results')\n$('#results').html(\"<p class='red'>"..result.."</p>\")\n$('#export').attr('exporting', 0)\n"
			else
				local sMsg =
					"<p class='green'>"..
					format("Processed : %d<br/>", result.total)..
					format("Success : %d<br/>", result.success)..
					format("Failed : %d<br/>", result.failed)..
					"</p>"
				local response = "toastr['success']('Success - see results')\n"
				if post.deck and post.deck ~= "" then
					if result.deckCopied then
						sMsg = sMsg.."<p class='green'>Deck copied successfully</p>"
						response = response .. "toastr['success']('Deck copied successfully')\n"
					else
						sMsg = sMsg.."<p class='red'>Error copying deck</p>"
						response = response .. "toastr['error']('Error copying deck')\n"
					end
				end
				if post.schedule and post.schedule ~= "" then
					if result.scheduleCopied then
						sMsg = sMsg.."<p class='green'>Schedule copied successfully</p>"
						response = response .. "toastr['success']('Schedule copied successfully')\n"
					else
						sMsg = sMsg.."<p class='red'>Error copying schedule</p>"
						response = response .. "toastr['error']('Error copying schedule')\n"
					end
				end
				if post.sfx then
					if result.sfxFailed == 0 then
						sMsg = sMsg.."<p class='green'>SFX copied successfully</p>"
						response = response .. "toastr['success']('SFX copied successfully')\n"
					else
						sMsg = sMsg..format("<p class='amber'>SFX copied with %d errors</p>", result.sfxFailed)
						response = response .. "toastr['warning']('"..format("SFX copied with %d errors", result.sfxFailed).."')\n"
					end
				end
				return response.."$('#results').html(\""..sMsg.."\")\n$('#export').attr('exporting', 0)\n"
			end
		elseif post.action == "files" then
			channel.main:push(json.encode({ id = "combobFiles" }))
			return "toastr['success']('Opening folder')\n"
		elseif post.action == "wait" then
			if appVars.combobData and appVars.combobData.action == "rewired" then
				local data = appVars.combobData
				local response = "" --"clearInterval(wait)\n"
				if data.success then
					response =
						"$('#result').html(\"<p class='green'>"..(data.msg or "").."</p>\")\n"..
						"$('#result').append(\"<p>You can import this file from the OBS menu : OBS / Scene Collection / Import</p>\")\n"
						if data.deck then
							if data.deck > 0 then
								response = 
									response..
									"$('#result').append(\"<br/><p class='green'>Deck file imported successfully : "..data.deckName.."</p>\")\n"..
									"$('#result').append(\"<p>You can restore this deck to the UP Deck app on your device</p>\")\n"
							else
								response = response.."$('#result').append(\"<br/><p class='red'>Error importing deck file</p>\")\n"
							end
						end
						if data.schedule then
							if data.schedule > 0 then
								response = response.."$('#result').append(\"<br/><p class='green'>Schedule file imported successfully : "..data.scheduleName.."</p>\")\n"
							else
								response = response.."$('#result').append(\"<br/><p class='red'>Error importing schedule file</p>\")\n"
							end
						end
						if data.sfx then
							response = response.."$('#result').append(\"<br/><p class='amber'>sfx folder detected - drag and drop it on top of the desktop app if you want to import it</p>\")\n"
						end
				else
					response = "$('#result').html(\"<p class='red'>Error : "..(data.msg or "").."</p>\")\n"
				end
				appVars.combobData = nil
				return response
			else
				return "console.log('Waiting...')\n"
			end
		end
	elseif post.id == "volume" then
		if post.action == "save" then
			local resp = ""

			-- update deck backup file
			local deck = uiVars[ip].deck
			if not deck then
				resp = resp.."toastr['error']('No deck is currently selected - deck file not updated')\n"
			else
				local data = fs.read("Backup/"..deck)
				if not data then
					resp = resp.."toastr['error']('Could not open deck file - not updated')\n"
				else
					local check
					check, data = pcall(jsonDecode, data)
					if check and type(data) == "table" and type(data.buttonData) == "table" and type(data.groupData) == "table" and type(data.sliderData) == "table" then
						-- create volume data
						local volData = {}
						for i = 1, 8 do
							volData[i] = post["v"..i] or ""
							if volData[i] == "" then volData[i] = "___null___" end
						end
						data.sliderData = volData
						-- save to file
						local success, msg = fs.write("Backup/"..deck, json.encode(data, "___null___"))
						if success then
							resp = resp.."toastr['success']('Changes saved to deck file')\n"
						else
							resp = resp.."toastr['error']('Could not write to deck file - not updated')\n"
						end
						-- update device
						uiVars[ip].client = post.client
						local client = uiVars[ip].client
						if not client or client == "" then
							resp = resp.."toastr['warning']('No device has been selected to update')\n"
						elseif not tcpClient[client] then
							resp = resp.."toastr['error']('Selected device is no longer connected')\n"
						else
							channel.tcp:push(json.encode({ id = "volupdate", data = {client = client, data = volData} }))
							resp = resp..format("toastr['success']('Changes sent to device : %s')\n", tcpClient[client] or "?")
						end
					else
						resp = resp.."toastr['error']('Invalid deck file - not updated')\n"
					end
				end
			end	

			-- return alerts to web UI
			resp = resp.."window.scrollTo(0,0)\n"
			return resp
		end
	elseif post.id == "util" then
		if post.action == "devices" then
			-- refresh connected devices
			local resp = "$('#client').empty().append(\"<option value=''>No device selected</option>"
			--local count = 1
			local selected
			local client = uiVars[ip].client or ""
			for k, v in pairs(tcpClient or {}) do
				--if count > 1 then resp = resp .. ", " end
				selected = client == k and "selected" or ""
				resp = resp..format(",<option %s value='%s'>%s</option>", selected, k, v)
				--count = count + 1
			end
			resp = resp.."\")\n"
			if (tonumber(post.noToast) or 0) ~= 1 then resp = resp.."toastr['success']('Device list updated')\n" end
			return resp
		elseif post.action == "client" then
			uiVars[ip].client = post.client
		elseif post.action == "decks" then
			-- refresh decks
			local backupFiles = {}
			local files = fs.getDirectoryItems("Backup")
			if files then
				for _, file in ipairs(files) do
					table.insert(backupFiles, file)
				end
				table.sort( backupFiles, function(a, b) return a:lower() < b:lower() end )
				local resp = "$('#deckSelect').empty().append(\""
				local count = 1
				local selected
				local deck = uiVars[ip].deck or ""
				for _, file in ipairs(backupFiles) do
					if count > 1 then resp = resp .. ", " end
					selected = deck == file and "selected" or ""
					resp = resp..format("<option %s value='%s'>%s</option>", selected, file, file)
					count = count + 1
				end
				resp = resp.."\")\n"
				return resp
			end
		elseif post.action == "schedules" then
			-- refresh schedules
			local schFiles = {}
			local files = fs.getDirectoryItems("Scheduler")
			if files then
				for _, file in ipairs(files) do
					table.insert(schFiles, file)
				end
				table.sort( schFiles, function(a, b) return a:lower() < b:lower() end )
				local resp = "$('#schSelect').empty().append(\""
				local count = 1
				local selected
				local schedule = uiVars[ip].schedule or ""
				for _, file in ipairs(schFiles) do
					if count > 1 then resp = resp .. ", " end
					selected = schedule == file and "selected" or ""
					resp = resp..format("<option %s value='%s'>%s</option>", selected, file, file:gsub("_", " "))
					count = count + 1
				end
				resp = resp.."\")\n"
				return resp
			end
		end
	elseif post.id == "login" and post.auth then
		local auth = love.data.encode("string", "hex", love.data.hash("sha1", cfg.webpass..ip))
		if auth == post.auth then
			saveUIVar(ip, "auth", true)
			return "window.location.href='/'\n"
		else
			return "toastr['error']('Incorrect password')\n$('#pass').val('')\n"
		end
	end
end


local function listen()
	if not server then
		local address = cfg.webopen and "*" or "127.0.0.1"
		local port = cfg.webport or 10314
		server = socket.bind(address, port, 0)
		if server then
			server:settimeout(0)
			server:setoption("reuseaddr", true)
			if cfg.debug then debug(format("Webserver listening on port %d", port)) end
		end
		return nil
	end

	-- check for request
	local c = server:accept()
	if c then
		local ip, _port = c:getpeername()
		local sock = ip..":".._port
		saveUIVar(ip, "time", floor(socket.gettime() * 1000))
		saveUIVar(ip, "ip", ip)

		--if cfg.debug then debug("Connected: "..sock) end
		c:settimeout(0)
		c:setoption("tcp-nodelay", true)
		c:setoption("keepalive", true)
	
		local req = nil
		local post = {}
		repeat
			local line, err = c:receive("*l")
			if not line then
				if err ~= "timeout" then
					c:close()
				end
			else
				if not req then req = "" else req = req.."\n" end
				req = req..line
			end
		until not line

		if req and c.send then
			c:settimeout(10)
			local procData
			local response
			local status
			local method, url = req:match("^(%a+)%s+/([^%s]*)")
			if url == "" then url = "setup" end
			if method and url and url:len() > 0 then
				if method == "POST" then
					local bound = req:match("Content%-Type:.-;%s*boundary=([^\n]+)")
					if bound then
						local _bound = bound:gsub("%-", "%%%-")
						local start = 1
						while true do
							local i, j = req:find("\n%-%-".._bound.."\n.-".._bound, start)
							if not i then break end
							local chunk = req:sub(i, j)
							local name, data = chunk:match("name=\"([^\"]+)\".-\n\n(.-)\n%-%-".._bound)
							if name and data then post[name] = data end
							start = i + 1
						end
					end
					procData = process(ip, post) or ""
				end
				
				if url == "ajax" then
					response = "HTTP/1.1 200 OK\n"..
						"Server: "..serverId.."\n"..
						"Content-Type: text/plain\n"..
						"Connection: keep-alive\n"..
						"Keep-Alive: timeout=5, max=1000\n"..
						"Content-Length: "..procData:len().."\n\n"..
						procData
				elseif ip == "127.0.0.1" and url:sub(1, 6) == "xdeck/" then
					local xdeck, button = url:match("^xdeck%/([%w_]+)%/(%d+)$")
					local cmd, error
					if not xdeck or not button then
						error = "incorrect URL format"
					else
						button = tostring(button)
						local deck = fs.read("Backup/"..xdeck)
						if not deck then
							error = "Could not read deck file : "..xdeck
						else
							local check
							check, deck = pcall(jsonDecode, deck)
							if check and type(deck) == "table" then
								if type(deck.buttonData) == "table" and type(deck.buttonData[button]) == "table" then
									cmd = deck.buttonData[button].cmd or ""
								end
							else
								error = "Could not decode deck file : "..xdeck
							end
						end
					end
					if error then
						response =
						"HTTP/1.1 400 ERROR\r\n"..
						"Server: "..serverId.."\r\n"..
						"Content-Type: text/plain\r\n"..
						"Connection: close\r\n\r\n"..
						"Error : "..error.."\r\n"
						if cfg.debug then debug(format("xdeck : Error = %s", error)) end
					else
						cmd = cmd or ""
						response =
						"HTTP/1.1 200 OK\r\n"..
						"Server: "..serverId.."\r\n"..
						"Content-Type: text/plain\r\n"..
						"Connection: close\r\n\r\n"..
						"OK\r\n"
						if cfg.debug then debug(format("xdeck : %s button #%d", xdeck, button)) end
						channel.tcp:push(json.encode({ id = "xdeck", data = cmd.."\n" }))
					end
				else
					if ip ~= "127.0.0.1" and not uiVar(ip, "auth") then
						url = "login"
					elseif ip == "127.0.0.1" and url:sub(1, 5) == "live/" then
						local targetDeck, gW, gH, gP, gS = url:match("^live%/([%w_]+)%/(%d+)%/(%d+)%/(%d+)%/(%d+)")
						url = "live"
						if targetDeck then
							saveUIVar(ip, "targetDeck", targetDeck)
							saveUIVar(ip, "liveGridW_"..targetDeck, gW)
							saveUIVar(ip, "liveGridH_"..targetDeck, gH)
							saveUIVar(ip, "liveBank_"..targetDeck, gP)
							saveUIVar(ip, "liveIconSz_"..targetDeck, gS)
						end
					elseif ip == "127.0.0.1" and url:sub(1, 9) == "liveDeck/" then
						local targetDeck = url:match("^liveDeck%/([%w_]+)$")
						url = "liveDeck"
						saveUIVar(ip, "targetDeck", targetDeck)
					end

					info = fs.getInfo("web/"..url)
					if not info then
						if cfg.debug then debug(format("%s : %s %s (404)", sock, method, url)) end
						status = "404"
						info = true
					end

					if info then
						if not status then status = "200" end
						local ext = url:match("[.](%w+)$")
						--if not ext and cfg.debug then debug(format("%s : %s %s", sock, method, url)) end
						if cfg.debug then debug(format("%s : %s %s", sock, method, url)) end
						local cType = (ext and mime[ext]) and mime[ext] or "text/html"
						local data = fs.read("web/"..url) or fs.read("web/404") or ""
						local cmd, param
						while true do
							cmd, param = data:match("%[%!([%u]+) ([^%]]+)%]")
							if not cmd or not param then
								break
							elseif cmd == "VERSION" then
								data = data:gsub("%[%!"..cmd.." "..param.."%]", version)
							elseif cmd == "INC" then
								-- include file
								data = data:gsub("%[%!"..cmd.." "..param.."%]", fs.read("web/"..param) or "")
							elseif cmd == "DO" then
								-- execut file
								local ok, chunk, result
								ok, chunk = pcall(fs.load, "web/"..param)
								if not ok then
									result = "Load Error: "..tostring(chunk)
								else
									local args = { url, uiVars[ip], appVars }
									ok, result = pcall(chunk, unpack(args))
									if not ok then
										result = "Chunk Error: "..tostring(result)
									else
										result = tostring(result)
									end
								end
								data = data:gsub("%[%!"..cmd.." "..param.."%]", result or "", 1)
							elseif cmd == "CFG" then
								-- updeck cfg value : text input
								data = data:gsub("%[%!"..cmd.." "..param.."%]", cfg[param] or "")
							elseif cmd == "CFGCHK" then
								-- updeck cfg value : checkbox
								data = data:gsub("%[%!"..cmd.." "..param.."%]", cfg[param] and "checked" or "")
							elseif cmd == "VAR" then
								-- fetch web var
								data = data:gsub("%[%!"..cmd.." "..param.."%]", uiVar(ip, param) or "", 1)
							elseif cmd == "SET" then
								-- set web var
								local k, v
								k, v = param:match("^([%a%d]+)=(.+)$")
								if k and v then saveUIVar(ip, k, v) end
								-- no output, just remove
								data = data:gsub("%[%!"..cmd.." "..param.."%]", "", 1)
							elseif cmd == "PNG" then
								local file = param
								if file:sub(1, 1) ~= "/" then file = "/"..file end
								file = "web"..file
								local b64Data = love.data.encode("string", "base64", fs.read(file) or "")
								data = data:gsub("%[%!"..cmd.." [^%]]+%]", function() return format("data:image/png;base64,%s", b64Data) end, 1)
							elseif cmd == "CSS" then
								local file = param
								if file:sub(1, 1) ~= "/" then file = "/"..file end
								file = "web"..file
								local cssData = "<style>\n"..(fs.read(file) or "").."\n</style>\n"
								data = data:gsub("%[%!"..cmd.." [^%]]+%]", function() return cssData end, 1)
							elseif cmd == "JS" then
								local file = param
								if file:sub(1, 1) ~= "/" then file = "/"..file end
								file = "web"..file
								local jsData = "<script>\n"..(fs.read(file) or "").."\n</script>\n"
								data = data:gsub("%[%!"..cmd.." [^%]]+%]", function() return jsData end, 1)
							else
								-- not recognized, remove from page
								data = data:gsub("%[%!"..cmd.." "..param.."%]", "")
							end
						end
						response = "HTTP/1.1 "..status.." OK\n"..
							"Server: "..serverId.."\n"..
							"Content-Type: "..cType.."\n"..
							"Connection: keep-alive\n"..
							"Keep-Alive: timeout=5, max=1000\n"..
							"Content-Length: "..data:len().."\n\n"..
							data
					else
						local data =
							"<html>\n<head><title>UPDeck Web Interface</title></head>\n"..
							"<body>Resource not found!</body>\n</html>\n"
						response = "HTTP/1.1 404 Not Found\n"..
							"Server: "..serverId.."\n"..
							"Content-Type: text/html\n"..
							"Connection: keep-alive\n"..
							"Keep-Alive: timeout=5, max=1000\n"..
							"Content-Length: "..data:len().."\n\n"..
							data
					end
				end
			else
				response = "HTTP/1.1 400 ERROR\n"..
					"Server: "..serverId.."\n"..
					"Content-Type: text/html\n"..
					"Connection: close\n\n"..
					"<html>\n<head><title>UPDeck Web Interface</title></head>\n"..
					"<body>Error</body>\n</html>\n"
			end
			c:send(response)
		end

		c:close()
	end -- request
end


if fs.isFused() then
    local dir = fs.getSourceBaseDirectory()
    local success = love.filesystem.mount(dir, "web")
    if not success then debug("Could not mount web folder!") end
end


while true do
	local chData
	while true do
		chData = channel.http:pop()
		if not chData then break end
		chData = json.decode(chData) or {}
		if not chData.id then break end
		if chData.id == "tcpclients" and chData.data then
			tcpClient = chData.data
		elseif chData.id == "schedule" then
			if chData.start then
				startSchedule(tostring(chData.start) or "")
			elseif chData.stop then
				stopSchedule()
			end
		elseif chData.id == "reset" then
			reset()
		elseif chData.id == "combob" then
			appVars.combobData = chData.data
		end
	end

	listen()

	if schJobs then
		local time = os.time()
		for k, v in pairs(schJobs) do
			if v.live and tonumber(v.target) and tonumber(v.interval) and v.target <= time then
				if cfg.debug then debug("Scheduler job: "..k) end
				channel.tcp:push(json.encode({ id = "schedule", data = v.cmd.."\n" }))
				v.target = time + v.interval
			end
		end
	end

	timer.sleep(0.02)
end