--[[


UP Deck
See version.lua


--]]
local json = require "lunajson"
local channel = require "channel"
local platform = love.system.getOS()
local tcpThread, httpThread
local connected = false
local cfg = {}
local fs = love.filesystem
local gr = love.graphics
local fpsCap = 20
local fInt = 1 / fpsCap
local topTile, topQuad, logo
local canvasData = {}
local canvasChData = nil
local canvasWidth, canvasHeight, canvasSize = 640, 360, 0.5
local canvasTimer, canvasTimerReset = 1, -8
local canvas = gr.newCanvas(
	canvasWidth,
	canvasHeight,
	{
		type = "2d",
		format = rgba8,
		readable = true,
		msaa = 0
	}
)


require "fileLib"


if fs.isFused() then
	local success = love.filesystem.mount(fs.getSourceBaseDirectory(), "_sourcebase")
	if not success then print("Could not mount base folder!") end
end


if platform == "Windows" then
	fs.rDelete("Win")
	if not fs.getInfo("Win") then fs.createDirectory("Win") end
	if not fs.getInfo("Win/Proc") then fs.createDirectory("Win/Proc") end
	local f, msg
	for _, script in pairs(fs.getDirectoryItems("winscripts")) do
		fs.remove(script)
		f, msg = fs.copyWinScript(script)
		if not f then print(string.format("Could not copy %s : %s", script, msg)) end
	end
end


if not fs.getInfo("Transitions") then
	fs.createDirectory("Transitions")
	fs.createDirectory("Transitions/Images")
	--[[
	local f
	for _, script in pairs(fs.getDirectoryItems("scrtrans")) do
		f = fs.read("scrtrans/"..script)
		if f then fs.write("Transitions/"..script, f) end
	end
	for _, image in pairs(fs.getDirectoryItems("scrtrans/images")) do
		f = fs.read("scrtrans/images/"..image)
		if f then fs.write("Transitions/Images/"..image, f) end
	end
	--]]
end


local function sfxFetch(path)
	if not path or path == "" then return nil end
	local sfxErr
	--fs.rDelete("Sounds")
	if not fs.getInfo("Sounds") then fs.createDirectory("Sounds") end
	if fs.getInfo("Sounds") and fs.getInfo("Sounds").type == "directory" then
		if fs.mount(path, "_sfx") then
			sfxErr = fs.rCopy("_sfx", "Sounds")
			fs.unmount(path)
			if sfxErr == 0 then
				love.window.showMessageBox("Sound Effects Import", "Sound effects imported successfully!", "info", false)
			else
				love.window.showMessageBox("Sound Effects Import", "An error occurred importing sound effects!\n(Could not copy files)", "error", false)
			end
		else
			sfxErr = -2
			if debug then print("Mount sfx failed:", path) end
			love.window.showMessageBox("Sound Effects Import", "An error occurred importing sound effects!", "error", false)
		end
	else
		sfxErr = -1
		if debug then print("Could not create Sounds folder!") end
		love.window.showMessageBox("Sound Effects Import", "An error occurred importing sound effects!\n(Could not create the folder)", "error", false)
	end
end


local function openAppFolder()
	love.system.openURL("file://"..fs.getSaveDirectory())
end


local function openCombobFolder()
	if not fs.getInfo("Combobulator") then
		fs.createDirectory("Combobulator")
		if not fs.getInfo("Combobulator") then
			love.window.showMessageBox("UP Deck", "Could not create Combobulator folder!", "error")
		end
	end
	if fs.getInfo("Combobulator") then
		love.system.openURL("file://"..fs.getSaveDirectory().."/Combobulator")
	end
end


local function openBackupFolder()
	if not fs.getInfo("Backup") then
		fs.createDirectory("Backup")
		if not fs.getInfo("Backup") then
			love.window.showMessageBox("UP Deck", "Could not create Backup folder!", "error")
		end
	end
	if fs.getInfo("Backup") then
		love.system.openURL("file://"..fs.getSaveDirectory().."/Backup")
	end
end


function love.load()
	topTile = gr.newImage("topTile.png")
	topTile:setWrap("repeat", "repeat")
	topQuad = gr.newQuad( 0, 0, gr.getWidth(), 76, 16, 76 )
	logo = gr.newImage("UPDeckLogo.png")

	cfg = {}
	if not fs.getInfo("updeck.txt") then
		local defCfg = {
			port = 4445,
			pass = "",
			tout = 30,
			webport = 10314,
			webpass = "",
			webopen = false,
			sfx = true,
			debug = false
		}
		fs.write("updeck.txt", json.encode(defCfg))
	end
	local data, msg = fs.read("updeck.txt")
	if not data then
		if debug then print("Could not read config", msg) end
	else
		cfg = json.decode(data) or {}
	end
	cfg.port = tonumber(cfg.port) or 4445
	if cfg.port < 1024 then cfg.port = 4445 end
	cfg.pass = cfg.pass or ""
	cfg.tout = tonumber(cfg.tout) or 0
	if cfg.tout < 10 then cfg.tout = 0 end
	cfg.sfx = cfg.sfx and true or false
	cfg.debug = cfg.debug and true or false
	cfg.webport = tonumber(cfg.webport) or 10314
	cfg.webpass = cfg.webpass or ""
	if cfg.webport < 1024 then cfg.webport = 10314 end
	cfg.webopen = cfg.webopen and true or false
	tcpThread = love.thread.newThread("tcpThread.lua")
	tcpThread:start(cfg)
	httpThread = love.thread.newThread("httpThread.lua")
	httpThread:start(cfg)
	if not fs.getInfo("Backup") then fs.createDirectory("Backup") end
	if not fs.getInfo("Lists") then fs.createDirectory("Lists") end
	if not fs.getInfo("Scheduler") then fs.createDirectory("Scheduler") end
end


function love.update(dt)
	if platform == "Windows" then
		-- process scripts
		local path = fs.getSaveDirectory()
		if path:sub(-1) ~= "/" then path = path.."/" end
		local scriptFiles = {}
		local files = fs.getDirectoryItems("Win")
		for _, file in ipairs(files) do
			table.insert(scriptFiles, file)
		end
		table.sort( scriptFiles, function(a, b) return a:lower() < b:lower() end )
		for _, file in ipairs(scriptFiles) do
			local info = fs.getInfo("Win/"..file)
			if info and info.type == "file" then
				local data = fs.read("Win/"..file)
				if data then
					if fs.remove("Win/"..file) and fs.write("Win/Proc/"..file, data) then
						local cl = 'start "" /B "'..string.gsub(path..'/Win/Proc/'..file, '/', '\\')..'"'
						os.execute(cl)
					end
				end
			end
		end
	end
	local chData
	while true do
		chData = channel.main:pop()
		if not chData then break end
		chData = json.decode(chData) or {}
		if not chData.id then break end
		if chData.id == "debug" and chData.data then
			print(chData.data)
		elseif chData.id == "cfg" and chData.data then
			cfg = chData.data
		elseif chData.id == "reset" then
			channel.tcp:push(json.encode({ id = "reset" }))
			channel.http:push(json.encode({ id = "reset" }))
		elseif chData.id == "files" then
			openAppFolder()
		elseif chData.id == "combobFiles" then
			openCombobFolder()
		elseif chData.id == "connection" then
			connected = (tonumber(chData.data or 0) > 0) and true or false
		elseif chData.id == "canvas" and chData.data then
			canvasChData = chData.data
		end
	end

	-- limit FPS
	local s = fInt - dt
	if s > fInt then s = fInt end
	if s > 0 then love.timer.sleep(s) end
end


function love.draw()
	gr.setColor(1, 1, 1)
	gr.draw(topTile, topQuad, 0, 0)
	gr.draw(logo, 8, 10)
	gr.setColor(0, 0, 0)
	gr.circle("fill", 205, 40, 10, 36)
	if connected then
		gr.setColor(0, 1, 0)
	else
		gr.setColor(1, 0, 0)
	end
	gr.circle("fill", 205, 40, 8, 36)

	if canvasChData and type(canvasChData) == "table" then
		local data = canvasChData
		canvasChData = nil
		if data.clear then
			canvasData = {}
		elseif data.undo then
			local undoCount = 0
			while data.undo and #canvasData > 0 do
				undoCount = undoCount + 1
				local last = table.remove(canvasData)
				if last.stop and undoCount > 1 then data.undo = nil end
			end
			if #canvasData > 0 then canvasData[#canvasData].stop = true end
		else
			if data.r and data.g and data.b then
				data.color = { data.r or 0, data.g or 0.8, data.b or 0 }
			end
			canvasData[#canvasData + 1] = data
		end

	
		local _x, _y
		gr.setCanvas(canvas)
		gr.setBlendMode("alpha")
		gr.setLineStyle("smooth")
		gr.clear(0, 0, 0, 0)
		gr.setColor(0, 0, 0, 0.01)
		gr.line(0, 0, 1, 1)
		for i, data in ipairs(canvasData) do
			if data.color then gr.setColor(unpack(data.color)) end
			if data.size then gr.setLineWidth(data.size) end
			if data.x and data.y then
				if _x and _y then gr.line(_x, _y, data.x, data.y) end
				_x, _y = data.x, data.y
			end
			if data.stop then _x, _y = nil, nil end
		end
		gr.setCanvas()
		canvas:newImageData():encode("png", "overlay.png")
	end
end


function love.mousepressed(x, y, button, istouch, presses)
	if button == 1 and presses > 1 then love.system.openURL("http://127.0.0.1:"..tonumber(cfg.webport or 10314)) end
end


function love.directorydropped(path)
	sfxFetch(path)
end


function love.filedropped(dropFile)
	local path = dropFile:getFilename():gsub("\\", "/")
	local file = path:match("[^/\\.]+%.-%w+$") or ""
	local ext = file:match("(%.%w+)$") or ""
	local newFile = file:sub(1, -ext:len() - 1).."_rewired"..ext
	local folder = path:sub(1, path:len() - file:len() - 1)
	local success = false
	local sfx = false
	local deck = false
	local deckName = ""
	local schedule = false
	local scheduleName = ""
	local mpSubs, cpSubs = 0, 0
	local msg = "File not recognised"
	if file ~= "" and ext == ".json" and folder:len() > 1 then
		local data = dropFile:read()
		if not data then
			msg = "Could not read file contents"
		else
			local ok, jData = pcall(json.decode, data)
			if ok and jData.name and type(jData.sources) == "table" then
				data, mpSubs = data:gsub("%[U_MESSAGE_PATH%]", fs.getSaveDirectory())
				data, cpSubs = data:gsub("%[U_COMBOBULATOR_PATH%]", folder)
				if mpSubs + cpSubs > 0 then
					local df = io.open(folder.."/"..newFile, "wb")
					if df and df:write(data) and df:close() then
						-- dropFile:open("w") and dropFile:write(data) and dropFile:close()
						success = true
						msg = "File successfully rewired : "..newFile
						-- check for deck file
						df = io.open(folder.."/deck.txt")
						if df then
							deck = 0
							local deckData = df:read("*a") or ""
							df:close()
							local dName, dData = deckData:match("^([^\n]+)\n(.+)$")
							if dName and dData then
								deckName = dName.."_"..os.time()
								if fs.write("Backup/"..deckName, dData) then deck = 1 end
							end
						end
						df = io.open(folder.."/schedule.txt")
						if df then
							schedule = 0
							local scheduleData = df:read("*a") or ""
							df:close()
							local sName, sData = scheduleData:match("^([^\n]+)\n(.+)$")
							if sName and sData then
								scheduleName = sName.."_"..os.time()
								if fs.write("Scheduler/"..scheduleName, sData) then schedule = 1 end
							end
						end
						df = io.open(folder.."/sfx")
						if df then
							sfx = true
							df:close()
						end
					else
						msg = "Could not write data to file"
					end
				else
					msg = "No changes were made to file"
				end
			else
				msg = "Data not in correct format"
			end
		end
	end
	channel.http:push(json.encode({
		id = "combob",
		data = {
			action = "rewired",
			success = success,
			newFile = newFile,
			mpSubs = mpSubs,
			cpSubs = cpSubs,
			rewired = mpSubs + cpSubs,
			msg = msg,
			deck = deck,
			deckName = deckName,
			schedule = schedule,
			scheduleName = scheduleName,
			sfx = sfx,
		}
	}))
end