-- Verion : 0.1
-- Released : 2019-07-11


function scriptTransition()


	-- Check the busy flag
	if scrSwitch.busy then
		obs.script_log(obs.LOG_INFO, "Transition already in progress")
		return false
	end
	scrSwitch.busy = true


	-- The scene to perform the transition in
	local transSceneName = "UPD_TRANSITION"


	-- The current and target scenes
	local sceneNameA = currentSceneName() or ""
	local sceneNameB = vParams.scene or ""
	if sceneNameB == "_previous" and prevScene then sceneNameB = prevScene end


	--[[
		This transition will create a copy of the current scene with a group on top
	 	containing the target scene and a set of 32 rectangles (boxes). The boxes
	 	completely cover the target scene and the group uses a chroma key filter
	 	so the boxes make the target scene invisible.  By rotating and shrinking the
	 	boxes, the target scene is revealed.
	--]]
	local groupName = "UPD_TRANS_GROUP"       -- group for target scene and boxes
	local maskName = "UPD_TRANS_OBJ"          -- prefix for the box objects
	local filterName = "UPD_TRANS_FILTER_CK"  -- chroma key filter applied to the group
	local baseW, baseH                        -- used for screen width / height


	-- Declare OBS objects here and use the releaseAll function to make sure
	-- we don't create any memory leaks
	local transScene, transSceneSource, group, sourceA, sourceB, filterCK
	local maskObj = {} -- array of source references to box objects
	local function releaseAll(animating)
		if not animating then scrSwitch.busy = false end
		if transScene then obs.obs_scene_release(transScene) end
		if transSceneSource then obs.obs_source_release(transSceneSource) end
		if group then obs.obs_source_release(group) end
		if sourceA then obs.obs_source_release(sourceA) end
		if sourceB then obs.obs_source_release(sourceB) end
		if filterCK then obs.obs_source_release(filterCK) end
		for _, obj in ipairs(maskObj) do obs.obs_source_release(obj) end
	end


	-- Abort the transition if the current scene or target scene is the special transition scene
	if sceneNameA == transSceneName or sceneNameB == transSceneName then
		obs.script_log(obs.LOG_INFO, format("Aborting - current scene : %s", transSceneName))
		releaseAll()
		return false
	end


	-- Get a reference to the transition scene and its source
	transSceneSource = obs.obs_get_source_by_name(transSceneName)
	if not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating scene : %s", transSceneName)) end
		transScene = obs.obs_scene_create(transSceneName)
		if transScene then transSceneSource = obs.obs_get_source_by_name(transSceneName) end
	else
		if debug then obs.script_log(obs.LOG_INFO, format("Found scene : %s", transSceneName)) end
		transScene = obs.obs_scene_from_source(transSceneSource)
		if transScene then obs.obs_scene_addref(transScene) end
	end
	if not transScene or not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Error accessing scene : %s", transSceneName)) end
		releaseAll()
		return false
	end


	-- Remove any objects in the transition scene
	emptyScene(transScene)


	-- Add the current scene to the transition scene
	sourceA = obs.obs_get_source_by_name(sceneNameA)
	if not sourceA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneA : %s", sceneNameA)) end
	local itemA = obs.obs_scene_add(transScene, sourceA)
	if not itemA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end


	-- Grab the screen dimensions
	baseW = obs.obs_source_get_width(sourceA)
	baseH = obs.obs_source_get_height(sourceA)
	if debug then obs.script_log(obs.LOG_INFO, format("Canvas : %d x %d", baseW, baseH)) end


	-- Create the group on top of the current scene
	group = obs.obs_get_source_by_name(groupName)
	if not group then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating group : %s", groupName)) end
		group = obs.obs_source_create("group", groupName, nil, nil)
		if not group then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't find group : %s", groupName)) end
			releaseAll()
			return false
		end
	end


	-- Create a chroma key filter to add to the group
	filterCK = obs.obs_get_source_by_name(filterName)
	if not filterCK then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating CK filter : %s", filterName)) end
		filterCK = obs.obs_source_create("chroma_key_filter", filterName, nil, nil)
		if not filterCK then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create CK filter : %s", filterName)) end
			releaseAll()
			return false
		end
	end


	-- This block of code sets the filter properties - tweak if required!
	-- Anything bright green in the group will show the current scene from below.
	local s = obs.obs_source_get_settings(filterCK)
	if s then
		obs.obs_data_set_string(s, "key_color_type", "custom")
		obs.obs_data_set_int(s, "key_color", 4278255360)
		obs.obs_data_set_int(s, "similarity", 256)
		obs.obs_data_set_int(s, "smoothness", 32)
		obs.obs_data_set_int(s, "spill", 64)
		obs.obs_data_set_int(s, "opacity", 100)
		obs.obs_source_update(filterCK, s)
		obs.obs_data_release(s)
	end


	-- Add the filter to the group
	local f = obs.obs_source_get_filter_by_name(group, filterName)
	if not f then
		if debug then obs.script_log(obs.LOG_INFO, format("Adding CK filter : %s", filterName)) end
		obs.obs_source_filter_add(group, filterCK)
	else
		obs.obs_source_release(f)
	end


	-- Add the group to the transition scene
	local groupItem = obs.obs_scene_add(transScene, group)
	if not groupItem then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add group to scene : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Get the group scene reference so we can add items to it
	local groupScene = obs.obs_sceneitem_group_get_scene(groupItem)
	if not groupScene then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't access group : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Just in case the group already existed - make sure it's empty
	emptyScene(groupScene)


	-- Now add the target scene to the group.  This will be invisible when we add
	-- green boxes on top (chroma key filter on group)
	sourceB = obs.obs_get_source_by_name(sceneNameB)
	if not sourceB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneB : %s", sceneNameB)) end
	local itemB = obs.obs_scene_add(groupScene, sourceB)
	if not itemB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end


	-- Create an 8 x 4 grid of green boxes starting at the top right
	-- and filling the screen with a snake pattern
	local cols = 8
	local rows = 4
	local objW = baseW / cols  -- width of box
	local objH = baseH / rows  -- height of box
	

	-- x,y vector for positioning boxes
	local pos = obs.vec2()
	pos.x, pos.y = baseW - objW * 0.5, objH * 0.5
	

	-- Size will be the same for all boxes
	local size = obs.vec2()
	size.x, size.y = objW, objH


	local d = 1  -- direction : 1 = down, -1 = up


	-- Create the boxes
	for i = 1, 32 do


		-- Create box number i
		local objName = "UPD_TRANS_OBJ_"..i
		maskObj[i] = obs.obs_get_source_by_name(objName)
		if not maskObj[i] then
			if debug then obs.script_log(obs.LOG_INFO, format("Creating mask object : %d", i)) end
			maskObj[i] = obs.obs_source_create("color_source", objName, nil, nil)
			if not maskObj[i] then
				if debug then obs.script_log(obs.LOG_INFO, format("Can't create mask object : %d", i)) end
				releaseAll()
				return false
			end
		end


		-- Set box width, height and color
		local s = obs.obs_source_get_settings(maskObj[i])
		if s then
			obs.obs_data_set_int(s, "width", objW)
			obs.obs_data_set_int(s, "height", objH)
			obs.obs_data_set_int(s, "color", 4278255360)
			obs.obs_source_update(maskObj[i], s)
			obs.obs_data_release(s)
		end


		-- Add the box to the group (on top of the target scene)
		local item = obs.obs_scene_add(groupScene, maskObj[i])
		if not item then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't add mask object to scene: %d", i)) end
			releaseAll()
			return false
		end


		-- Set the box position and size
		obs.obs_sceneitem_set_alignment(item, obsAlignment.center)
		obs.obs_sceneitem_set_bounds_type(item, obsBoundsType.stretch)
		obs.obs_sceneitem_set_bounds(item, size) -- w / h
		obs.obs_sceneitem_set_pos(item, pos) -- x / y
		obs.obs_sceneitem_set_rot(item, 0) -- rotation
		obs.obs_sceneitem_set_visible(item, true)


		-- Set the position / direction for creating the next box (snake)
		if i % rows == 0 then
			d = d * -1
			pos.x = pos.x - objW
		else
			pos.y = pos.y + objH * d
		end
	end


	-- Select the required transition to switch to our newly created
	-- transition scene - 'Cut' is the default.
	local transitions = obs.obs_frontend_get_transitions()
	local transition = findSource(transitions, vParams.trans_before or "Cut")
	if transition then obs.obs_frontend_set_current_transition(transition) end
	obs.source_list_release(transitions)


	-- record the current scene so that it can be referred to as '_previous'
	prevScene = currentSceneName()


	-- This outer timer block is run in 200 ms - you'll see the number near the bottom, below
	obs.timer_add(


		function()
			-- Switch to the transition scene
			obs.obs_frontend_set_current_scene(transSceneSource)


			obs.script_log(obs.LOG_INFO, "Transition ready to animate")


			-- Build a list of animate commands to rotate and shrink the boxes
			-- Each one with a slightly longer delay for the left to right effect
			-- Note the command format : no description and new lines replaced with tabs ( \t )
			-- Also a newline character ( \n ) at the end of each command
			local cmd = ""
			for i = 1, 32 do
				cmd = cmd..format(
					"animate\ttag=UPDtrans\tdelay=%d\tscene=%s\titem=%s\tgroup=%s\tsteps=50\tinterval=10\tr=(-90)\tw=0\th=0\n",
					40 * i, transSceneName, "UPD_TRANS_OBJ_"..i, groupName
				)
			end


			-- Send the commands - just as if they had been received from the mobile app.
			-- This will start the animations
			listen(cmd)


			-- Inner timer : run every 500 ms to check for completion
			obs.timer_add(


				function()


					-- Check if animations are still running
					if (animTag["UPDtrans"] or 0) > 0 then


						-- still busy
						obs.script_log(obs.LOG_INFO, "Transition in progress...")


					else


						-- Select the required transition to switch to the target scene
						-- from the transition scene - 'Cut' is the default.
						local transitions = obs.obs_frontend_get_transitions()
						local transition = findSource(transitions, vParams.trans_after or "Cut")
						if transition then obs.obs_frontend_set_current_transition(transition) end
						obs.source_list_release(transitions)


						-- The final switch to the target scene
						local targetScene = obs.obs_get_source_by_name(sceneNameB)
						if targetScene then
							obs.obs_frontend_set_current_scene(targetScene)
							obs.obs_source_release(targetScene)
						end


						-- Unblock new transitions
						scrSwitch.busy = false


						-- We are done! - remove the timer
						obs.script_log(obs.LOG_INFO, "Transition complete")
						obs.remove_current_callback()


					end
				end,
				500  -- (Inner timer runs every 500 ms)


			)  -- Inner timer


			-- Release all our OBS references, but use the 'animating' flag
			-- to let it know we are NOT finished
			releaseAll("animating")


			-- This block only needs to run once to get the animation moving, so remove it
			obs.remove_current_callback()
		end,
		200  -- (Outer timer runs after 200 ms)


	)  -- Outer timer


end