-- Verion : 1.0
-- Released : 2019-07-09


function scriptTransition()


	-- Check the busy flag
	if scrSwitch.busy then
		obs.script_log(obs.LOG_INFO, "Transition already in progress")
		return false
	end
	scrSwitch.busy = true


	-- The scene to perform the transition in
	local transSceneName = "UPD_TRANSITION"


	-- The current and target scenes
	local sceneNameA = currentSceneName() or ""
	local sceneNameB = vParams.scene or ""
	if sceneNameB == "_previous" and prevScene then sceneNameB = prevScene end


	--[[
		This transition will create a copy of the target scene with a group on top
	 	containing the current scene and the shapes to create see-through areas by
	 	using a chroma key filter applied to the group.
	--]]
	local groupName = "UPD_TRANS_GROUP"               -- group for current scene and shapes
	local maskName = "UPD_TRANS_OBJ"                  -- prefix for the shape objects
	local filterCKName = "UPD_TRANS_FILTER_CK"        -- chroma key filter applied to the group
	local filterCCName = "UP DECK COLOR"              -- color correction filter to fade current scene
	local filterSplatName = "UPD_TRANS_FILTER_SPLAT"  -- mask filter with a splat shape
	local baseW, baseH                                -- used for screen width / height


	-- Declare OBS objects here and use the releaseAll function to make sure
	-- we don't create any memory leaks
	local transScene, transSceneSource, group, sourceA, sourceB, filterCK, filterCC, filterSplat
	local maskObj = {} -- array of source references to shape objects
	local function releaseAll(animating)
		-- If the 'animating' flag is not set then we are finished and we can
		-- reset the busy flag to allow more transitions to be initiated
		if not animating then scrSwitch.busy = false end
		if transScene then obs.obs_scene_release(transScene) end
		if transSceneSource then obs.obs_source_release(transSceneSource) end
		if group then obs.obs_source_release(group) end
		if sourceA then obs.obs_source_release(sourceA) end
		if sourceB then obs.obs_source_release(sourceB) end
		if filterCK then obs.obs_source_release(filterCK) end
		if filterCC then obs.obs_source_release(filterCC) end
		if filterSplat then obs.obs_source_release(filterSplat) end
		for _, obj in ipairs(maskObj) do obs.obs_source_release(obj) end
	end


	-- Abort the transition if the current scene or target scene is the special transition scene
	if sceneNameA == transSceneName or sceneNameB == transSceneName then
		obs.script_log(obs.LOG_INFO, format("Aborting - current scene : %s", transSceneName))
		return false
	end


	-- Get a reference to the transition scene and its source
	transSceneSource = obs.obs_get_source_by_name(transSceneName)
	if not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating scene : %s", transSceneName)) end
		transScene = obs.obs_scene_create(transSceneName)
		if transScene then transSceneSource = obs.obs_get_source_by_name(transSceneName) end
	else
		if debug then obs.script_log(obs.LOG_INFO, format("Found scene : %s", transSceneName)) end
		transScene = obs.obs_scene_from_source(transSceneSource)
		if transScene then obs.obs_scene_addref(transScene) end
	end
	if not transScene or not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Error accessing scene : %s", transSceneName)) end
		releaseAll()
		return false
	end


	-- Remove any objects in the transition scene
	emptyScene(transScene)


	-- Add the target scene to the transition scene
	sourceB = obs.obs_get_source_by_name(sceneNameB)
	if not sourceB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneB : %s", sceneNameB)) end
	local itemB = obs.obs_scene_add(transScene, sourceB)
	if not itemB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end


	-- Grab the screen dimensions
	baseW = obs.obs_source_get_width(sourceB)
	baseH = obs.obs_source_get_height(sourceB)
	if debug then obs.script_log(obs.LOG_INFO, format("Canvas : %d x %d", baseW, baseH)) end


	-- Create the group on top of the target scene
	group = obs.obs_get_source_by_name(groupName)
	if not group then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating group : %s", groupName)) end
		group = obs.obs_source_create("group", groupName, nil, nil)
		if not group then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't find group : %s", groupName)) end
			releaseAll()
			return false
		end
	end


	-- Create a chroma key filter to add to the group
	filterCK = obs.obs_get_source_by_name(filterCKName)
	if not filterCK then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating CK filter : %s", filterCKName)) end
		filterCK = obs.obs_source_create("chroma_key_filter", filterCKName, nil, nil)
		if not filterCK then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create CK filter : %s", filterCKName)) end
			releaseAll()
			return false
		end
	end


	-- This block of code sets the filter properties - tweak if required!
	-- Anything bright green in the group will show the target scene below.
	local s = obs.obs_source_get_settings(filterCK)
	if s then
		obs.obs_data_set_string(s, "key_color_type", "custom")
		obs.obs_data_set_int(s, "key_color", 4278255360)
		obs.obs_data_set_int(s, "similarity", 256)
		obs.obs_data_set_int(s, "smoothness", 32)
		obs.obs_data_set_int(s, "spill", 64)
		obs.obs_data_set_int(s, "opacity", 100)
		obs.obs_source_update(filterCK, s)
		obs.obs_data_release(s)
	end


	-- Add the filter to the group
	local f = obs.obs_source_get_filter_by_name(group, filterCKName)
	if not f then
		if debug then obs.script_log(obs.LOG_INFO, format("Adding CK filter : %s", filterCKName)) end
		obs.obs_source_filter_add(group, filterCK)
	else
		obs.obs_source_release(f)
	end


	-- Add the group to the transition scene
	local groupItem = obs.obs_scene_add(transScene, group)
	if not groupItem then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add group to scene : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Get the group scene reference so we can add items to it
	local groupScene = obs.obs_sceneitem_group_get_scene(groupItem)
	if not groupScene then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't access group : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Just in case the group already existed - make sure it's empty
	emptyScene(groupScene)


	-- Now add the current scene to the group.  This will cover the target scene
	-- until the green shapes reveal it (chroma key filter on group)
	sourceA = obs.obs_get_source_by_name(sceneNameA)
	if not sourceA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneA : %s", sceneNameA)) end
	local itemA = obs.obs_scene_add(groupScene, sourceA)
	if not itemA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end


	-- Create a color correction filter to fade the current scene at the end
	filterCC = obs.obs_get_source_by_name(filterCCName)
	if not filterCC then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating CC filter : %s", filterCCName)) end
		filterCC = obs.obs_source_create("color_filter", filterCCName, nil, nil)
		if not filterCC then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create CC filter : %s", filterCCName)) end
			releaseAll()
			return false
		end
	end


	-- Set color correction filter to full opacity
	local s = obs.obs_source_get_settings(filterCC)
	if s then
		obs.obs_data_set_int(s, "opacity", 100)
		obs.obs_source_update(filterCC, s)
		obs.obs_data_release(s)
	end


	-- Add the filter to the current scene
	local f = obs.obs_source_get_filter_by_name(sourceA, filterCCName)
	if not f then
		if debug then obs.script_log(obs.LOG_INFO, format("Adding CC filter : %s", filterCCName)) end
		obs.obs_source_filter_add(sourceA, filterCC)
	else
		obs.obs_source_release(f)
	end


	-- Create a mask filter to change the shapes from plain rectangles
	filterSplat = obs.obs_get_source_by_name(filterSplatName)
	if not filterSplat then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating mask filter : %s", filterSplatName)) end
		filterSplat = obs.obs_source_create("mask_filter", filterSplatName, nil, nil)
		if not filterSplat then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create mask filter : %s", filterSplatName)) end
			releaseAll()
			return false
		end
	end


	-- Set mask filter filename to splat image
	local s = obs.obs_source_get_settings(filterSplat)
	if s then
		obs.obs_data_set_string(s, "image_path", msgPath.."/Transitions/Images/splat1.png")
		obs.obs_data_set_bool(s, "stretch", true)
		obs.obs_source_update(filterSplat, s)
		obs.obs_data_release(s)
	end


	-- x,y vector for positioning shapes
	local pos = obs.vec2()


	-- For setting shape sizes
	local size = obs.vec2()


	-- Create some shapes to reveal the target scene under the current scene
	-- They will start small and splat on to the screen
	local shapes = {
		{
			-- shape to reveal target scene
			x = baseW * 0.6,
			y = baseH * 0.1,
			width = 4,
			height = 12,
			targetWidth = baseH * 0.6,
			targetHeight = baseH * 0.7,
			delay = 0,
			steps = 10,
			interval = 10,
		},
		--[[
		{
			-- shape to reveal target scene
			x = baseW * 0.2,
			y = baseH * 0.3,
			width = 4,
			height = 12,
			targetWidth = baseH * 0.4,
			targetHeight = baseH * 0.4,
			delay = 50,
			steps = 10,
			interval = 10,
		},
		--[[
		{
			-- shape to reveal target scene
			x = baseW * 0.75,
			y = baseH * 0.4,
			width = 4,
			height = 12,
			targetWidth = baseH * 0.45,
			targetHeight = baseH * 0.6,
			delay = 75,
			steps = 10,
			interval = 10,
		},
		--]]
	}


	-- Create the shapes
	for i = 1, #shapes do


		-- Create shape number i
		local objName = maskName..i
		maskObj[i] = obs.obs_get_source_by_name(objName)
		if not maskObj[i] then
			if debug then obs.script_log(obs.LOG_INFO, format("Creating mask object : %d", i)) end
			maskObj[i] = obs.obs_source_create("color_source", objName, nil, nil)
			if not maskObj[i] then
				if debug then obs.script_log(obs.LOG_INFO, format("Can't create mask object : %d", i)) end
				releaseAll()
				return false
			end
		end


		-- Set shape width, height and color
		local s = obs.obs_source_get_settings(maskObj[i])
		if s then
			obs.obs_data_set_int(s, "width", 256)
			obs.obs_data_set_int(s, "height", 256)
			obs.obs_data_set_int(s, "color", 4278255360)
			obs.obs_source_update(maskObj[i], s)
			obs.obs_data_release(s)
		end


		-- Add the shape to the group (on top of the current scene)
		local item = obs.obs_scene_add(groupScene, maskObj[i])
		if not item then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't add mask object to scene: %d", i)) end
			releaseAll()
			return false
		end


		-- Set the shape position and size
		pos.x, pos.y = shapes[i].x, shapes[i].y
		size.x, size.y = shapes[i].width, shapes[i].height
		obs.obs_sceneitem_set_alignment(item, obsAlignment.top + obsAlignment.center)
		obs.obs_sceneitem_set_bounds_type(item, obsBoundsType.stretch)
		obs.obs_sceneitem_set_bounds(item, size) -- w / h
		obs.obs_sceneitem_set_pos(item, pos) -- x / y
		obs.obs_sceneitem_set_rot(item, 0) -- rotation
		obs.obs_sceneitem_set_visible(item, true)


		-- Add the mask filter to create the shape effect
		obs.obs_source_filter_add(maskObj[i], filterSplat)


	end


	-- Select the required transition to switch to our newly created
	-- transition scene - 'Cut' is the default.
	local transitions = obs.obs_frontend_get_transitions()
	local transition = findSource(transitions, vParams.trans_before or "Cut")
	if transition then obs.obs_frontend_set_current_transition(transition) end
	obs.source_list_release(transitions)


	-- record the current scene so that it can be referred to as '_previous'
	prevScene = currentSceneName()


	-- This outer timer block is run in 200 ms - you'll see the number near the bottom, below
	obs.timer_add(


		function()
			-- Switch to the transition scene
			obs.obs_frontend_set_current_scene(transSceneSource)


			obs.script_log(obs.LOG_INFO, "Transition ready to animate")


			-- Build a list of animate commands to create the splat effects
			-- Note the command format : no description and new lines replaced with tabs ( \t )
			-- Also a newline character ( \n ) at the end of each command
			local cmd = ""
			for i = 1, #shapes do
				cmd = cmd..
					format(
						"animate\ttag=UPDtrans\tqueue=1\tdelay=%d\tscene=%s\titem=%s\tgroup=%s\t"..
						"steps=%d\tinterval=%d\tw=%d\th=%d\teasing=EaseOutElastic\n",
						shapes[i].delay or 0, transSceneName, maskName..i, groupName,
						shapes[i].steps or 1, shapes[i].interval or 10,
						shapes[i].targetWidth, shapes[i].targetHeight
					)..
					format(
						"animate\ttag=UPDtrans\tscene=%s\titem=%s\tgroup=%s\t"..
						"steps=%d\tinterval=%d\ty=(%d)\tw=(%d)\th=(%d)\n",
						transSceneName, maskName..i, groupName, 100, 30,
						shapes[i].targetHeight * 0.1,
						shapes[i].targetWidth * 0.1,
						shapes[i].targetHeight * 0.25
					)
			end


			-- Send the commands - just as if they had been received from the mobile app.
			-- This will start the animations
			listen(cmd)


			-- Inner timer : run every 500 ms to check for completion
			obs.timer_add(


				function()


					-- Check if animations are still running
					if (animTag["UPDtrans"] or 0) > 0 then


						-- still busy
						obs.script_log(obs.LOG_INFO, "Transition in progress...")


					else


						-- Select the required transition to switch to the target scene
						-- from the transition scene - 'Cut' is the default.
						local transitions = obs.obs_frontend_get_transitions()
						local transition = findSource(transitions, vParams.trans_after or "Cut")
						if transition then obs.obs_frontend_set_current_transition(transition) end
						obs.source_list_release(transitions)


						-- The final switch to the target scene
						local targetScene = obs.obs_get_source_by_name(sceneNameB)
						if targetScene then
							obs.obs_frontend_set_current_scene(targetScene)
							obs.obs_source_release(targetScene)
						end


						-- Unblock new transitions
						scrSwitch.busy = false


						-- We are done! - remove the timer
						obs.script_log(obs.LOG_INFO, "Transition complete")
						obs.remove_current_callback()


					end
				end,
				500  -- (Inner timer runs every 500 ms)


			)  -- Inner timer


			-- Release all our OBS references, but use the 'animating' flag
			-- to let it know we are NOT finished
			releaseAll("animating")


			-- This block only needs to run once to get the animation moving, so remove it
			obs.remove_current_callback()
		end,
		200  -- (Outer timer runs after 200 ms)


	)  -- Outer timer


end