-- Verion : 0.1
-- Released : 2019-07-11


function scriptTransition()


	-- Check the busy flag
	if scrSwitch.busy then
		obs.script_log(obs.LOG_INFO, "Transition already in progress")
		return false
	end
	scrSwitch.busy = true


	-- The scene to perform the transition in
	local transSceneName = "UPD_TRANSITION"


	-- The current and target scenes
	local sceneNameA = currentSceneName() or ""
	local sceneNameB = vParams.scene or ""
	if sceneNameB == "_previous" and prevScene then sceneNameB = prevScene end


	--[[
		This transition will create a copy of the target scene with a group on top
	 	containing the current scene and a shape to create a see-through area by
	 	using a chroma key filter applied to the group.
	--]]
	local groupName = "UPD_TRANS_GROUP"               -- group for current scene and shape
	local maskName = "UPD_TRANS_OBJ"                  -- prefix for the shape objects
	local filterCKName = "UPD_TRANS_FILTER_CK"        -- chroma key filter applied to the group
	local filterSplatName = "UPD_TRANS_FILTER_SPLAT"  -- mask filter with a splat shape
	local baseW, baseH                                -- used for screen width / height


	-- Declare OBS objects here and use the releaseAll function to make sure
	-- we don't create any memory leaks
	local transScene, transSceneSource, group, sourceA, sourceB, filterCK, filterSplat, maskObj
	local function releaseAll(animating)
		-- If the 'animating' flag is not set then we are finished and we can
		-- reset the busy flag to allow more transitions to be initiated
		if not animating then scrSwitch.busy = false end
		if transScene then obs.obs_scene_release(transScene) end
		if transSceneSource then obs.obs_source_release(transSceneSource) end
		if group then obs.obs_source_release(group) end
		if sourceA then obs.obs_source_release(sourceA) end
		if sourceB then obs.obs_source_release(sourceB) end
		if filterCK then obs.obs_source_release(filterCK) end
		if filterSplat then obs.obs_source_release(filterSplat) end
		if maskObj then obs.obs_source_release(maskObj) end
	end


	-- Abort the transition if the current scene or target scene is the special transition scene
	if sceneNameA == transSceneName or sceneNameB == transSceneName then
		obs.script_log(obs.LOG_INFO, format("Aborting - current scene : %s", transSceneName))
		releaseAll()
		return false
	end


	-- Get a reference to the transition scene and its source
	transSceneSource = obs.obs_get_source_by_name(transSceneName)
	if not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating scene : %s", transSceneName)) end
		transScene = obs.obs_scene_create(transSceneName)
		if transScene then transSceneSource = obs.obs_get_source_by_name(transSceneName) end
	else
		if debug then obs.script_log(obs.LOG_INFO, format("Found scene : %s", transSceneName)) end
		transScene = obs.obs_scene_from_source(transSceneSource)
		if transScene then obs.obs_scene_addref(transScene) end
	end
	if not transScene or not transSceneSource then
		if debug then obs.script_log(obs.LOG_INFO, format("Error accessing scene : %s", transSceneName)) end
		releaseAll()
		return false
	end


	-- Remove any objects in the transition scene
	emptyScene(transScene)


	-- Add the target scene to the transition scene
	sourceB = obs.obs_get_source_by_name(sceneNameB)
	if not sourceB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneB : %s", sceneNameB)) end
	local itemB = obs.obs_scene_add(transScene, sourceB)
	if not itemB then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneB : %s", sceneNameB)) end
		releaseAll()
		return false
	end


	-- Grab the screen dimensions
	baseW = obs.obs_source_get_width(sourceB)
	baseH = obs.obs_source_get_height(sourceB)
	if debug then obs.script_log(obs.LOG_INFO, format("Canvas : %d x %d", baseW, baseH)) end


	-- Create the group on top of the target scene
	group = obs.obs_get_source_by_name(groupName)
	if not group then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating group : %s", groupName)) end
		group = obs.obs_source_create("group", groupName, nil, nil)
		if not group then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't find group : %s", groupName)) end
			releaseAll()
			return false
		end
	end


	-- Create a chroma key filter to add to the group
	filterCK = obs.obs_get_source_by_name(filterCKName)
	if not filterCK then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating CK filter : %s", filterCKName)) end
		filterCK = obs.obs_source_create("chroma_key_filter", filterCKName, nil, nil)
		if not filterCK then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create CK filter : %s", filterCKName)) end
			releaseAll()
			return false
		end
	end


	-- This block of code sets the filter properties - tweak if required!
	-- Anything bright green in the group will show the target scene below.
	local s = obs.obs_source_get_settings(filterCK)
	if s then
		obs.obs_data_set_string(s, "key_color_type", "custom")
		obs.obs_data_set_int(s, "key_color", 4278255360)
		obs.obs_data_set_int(s, "similarity", 320)
		obs.obs_data_set_int(s, "smoothness", 32)
		obs.obs_data_set_int(s, "spill", 64)
		obs.obs_data_set_int(s, "opacity", 100)
		obs.obs_source_update(filterCK, s)
		obs.obs_data_release(s)
	end


	-- Add the filter to the group
	local f = obs.obs_source_get_filter_by_name(group, filterCKName)
	if not f then
		if debug then obs.script_log(obs.LOG_INFO, format("Adding CK filter : %s", filterCKName)) end
		obs.obs_source_filter_add(group, filterCK)
	else
		obs.obs_source_release(f)
	end


	-- Add the group to the transition scene
	local groupItem = obs.obs_scene_add(transScene, group)
	if not groupItem then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add group to scene : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Get the group scene reference so we can add items to it
	local groupScene = obs.obs_sceneitem_group_get_scene(groupItem)
	if not groupScene then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't access group : %s", groupName)) end
		releaseAll()
		return false
	end


	-- Just in case the group already existed - make sure it's empty
	emptyScene(groupScene)


	-- Now add the current scene to the group.  This will cover the target scene
	-- until the green shape reveals it (chroma key filter on group)
	sourceA = obs.obs_get_source_by_name(sceneNameA)
	if not sourceA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't find sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end
	if debug then obs.script_log(obs.LOG_INFO, format("Adding sceneA : %s", sceneNameA)) end
	local itemA = obs.obs_scene_add(groupScene, sourceA)
	if not itemA then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add sceneA : %s", sceneNameA)) end
		releaseAll()
		return false
	end


	-- Create a mask filter to change the shape from a plain rectangle
	filterSplat = obs.obs_get_source_by_name(filterSplatName)
	if not filterSplat then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating mask filter : %s", filterSplatName)) end
		filterSplat = obs.obs_source_create("mask_filter", filterSplatName, nil, nil)
		if not filterSplat then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create mask filter : %s", filterSplatName)) end
			releaseAll()
			return false
		end
	end


	-- Set mask filter filename to splat image
	local s = obs.obs_source_get_settings(filterSplat)
	if s then
		obs.obs_data_set_string(s, "image_path", msgPath.."/Transitions/Images/splat1.png")
		obs.obs_data_set_bool(s, "stretch", true)
		obs.obs_source_update(filterSplat, s)
		obs.obs_data_release(s)
	end


	-- x,y vector for positioning shape
	local pos = obs.vec2()


	-- For setting shape size
	local size = obs.vec2()


	-- Create splat shape to reveal the target scene under the current scene
	-- It will start small, splat on to the screen then expand
	local shape = {
		x = baseW * 0.5,
		y = baseH * 0.5,
		width = 12,
		height = 12,
		targetX = baseW * 0.5,
		targetY = baseH * 0.5,
		targetWidth = baseW * 0.9,
		targetHeight = baseH * 0.9,
		delay = 0,
		steps = 8,
		interval = 8,
	}
	maskObj = obs.obs_get_source_by_name(maskName)
	if not maskObj then
		if debug then obs.script_log(obs.LOG_INFO, format("Creating mask object : %s", maskName)) end
		maskObj = obs.obs_source_create("color_source", maskName, nil, nil)
		if not maskObj then
			if debug then obs.script_log(obs.LOG_INFO, format("Can't create mask object : %s", maskName)) end
			releaseAll()
			return false
		end
	end


	-- Set shape width, height and color
	local s = obs.obs_source_get_settings(maskObj)
	if s then
		obs.obs_data_set_int(s, "width", 256)
		obs.obs_data_set_int(s, "height", 256)
		obs.obs_data_set_int(s, "color", 4278255360)
		obs.obs_source_update(maskObj, s)
		obs.obs_data_release(s)
	end


	-- Add the shape to the group (on top of the current scene)
	local maskItem = obs.obs_scene_add(groupScene, maskObj)
	if not maskItem then
		if debug then obs.script_log(obs.LOG_INFO, format("Can't add mask object to scene : %s", maskName)) end
		releaseAll()
		return false
	end


	-- Set the shape position and size
	pos.x, pos.y = shape.x, shape.y
	size.x, size.y = shape.width, shape.height
	obs.obs_sceneitem_set_alignment(maskItem, obsAlignment.center)
	obs.obs_sceneitem_set_bounds_type(maskItem, obsBoundsType.stretch)
	obs.obs_sceneitem_set_bounds(maskItem, size) -- w / h
	obs.obs_sceneitem_set_pos(maskItem, pos) -- x / y
	obs.obs_sceneitem_set_rot(maskItem, 0) -- rotation
	obs.obs_sceneitem_set_visible(maskItem, true)


	-- Add the mask filter to create the shape effect
	local f = obs.obs_source_get_filter_by_name(maskObj, filterSplatName)
	if not f then
		if debug then obs.script_log(obs.LOG_INFO, format("Adding mask filter : %s", filterSplatName)) end
		obs.obs_source_filter_add(maskObj, filterSplat)
	else
		obs.obs_source_release(f)
	end


	-- Select the required transition to switch to our newly created
	-- transition scene - 'Cut' is the default.
	local transitions = obs.obs_frontend_get_transitions()
	local transition = findSource(transitions, vParams.trans_before or "Cut")
	if transition then obs.obs_frontend_set_current_transition(transition) end
	obs.source_list_release(transitions)


	-- record the current scene so that it can be referred to as '_previous'
	prevScene = currentSceneName()


	-- This outer timer block is run in 200 ms - you'll see the number near the bottom, below
	obs.timer_add(


		function()
			-- Switch to the transition scene
			obs.obs_frontend_set_current_scene(transSceneSource)


			obs.script_log(obs.LOG_INFO, "Transition ready to animate")


			-- Build a list of animate commands to create the splat effects
			-- Note the command format : no description and new lines replaced with tabs ( \t )
			-- Also a newline character ( \n ) at the end of each command
			local cmd = 
				format(
					"animate\ttag=UPDtrans\tqueue=1\tdelay=%d\tscene=%s\titem=%s\tgroup=%s\t"..
					"x=%d\ty=%d\tsteps=%d\tinterval=%d\tw=%d\th=%d\teasing=EaseOut\n",
					shape.delay or 0, transSceneName, maskName, groupName,
					shape.targetX, shape.targetY,
					shape.steps or 1, shape.interval or 10,
					shape.targetWidth, shape.targetHeight
				)..
				format(
					"animate\ttag=UPDtrans\tscene=%s\titem=%s\tgroup=%s\t"..
					"steps=%d\tinterval=%d\tw=%d\th=%d\teasing=EaseIn\n",
					transSceneName, maskName, groupName, 50, 20,
					baseW * 2.5, baseH * 2.5
				)


			-- Send the commands - just as if they had been received from the mobile app.
			-- This will start the animations
			listen(cmd)


			-- Inner timer : run every 500 ms to check for completion
			obs.timer_add(


				function()


					-- Check if animations are still running
					if (animTag["UPDtrans"] or 0) > 0 then


						-- still busy
						obs.script_log(obs.LOG_INFO, "Transition in progress...")


					else


						-- Select the required transition to switch to the target scene
						-- from the transition scene - 'Cut' is the default.
						local transitions = obs.obs_frontend_get_transitions()
						local transition = findSource(transitions, vParams.trans_after or "Cut")
						if transition then obs.obs_frontend_set_current_transition(transition) end
						obs.source_list_release(transitions)


						-- The final switch to the target scene
						local targetScene = obs.obs_get_source_by_name(sceneNameB)
						if targetScene then
							obs.obs_frontend_set_current_scene(targetScene)
							obs.obs_source_release(targetScene)
						end


						-- Unblock new transitions
						scrSwitch.busy = false


						-- We are done! - remove the timer
						obs.script_log(obs.LOG_INFO, "Transition complete")
						obs.remove_current_callback()


					end
				end,
				500  -- (Inner timer runs every 500 ms)


			)  -- Inner timer


			-- Release all our OBS references, but use the 'animating' flag
			-- to let it know we are NOT finished
			releaseAll("animating")


			-- This block only needs to run once to get the animation moving, so remove it
			obs.remove_current_callback()
		end,
		200  -- (Outer timer runs after 200 ms)


	)  -- Outer timer


end