local socket = require "socket"
local system = require "love.system"
local timer = require "love.timer"
local audio = require "love.audio"
local fs = require "love.filesystem"
local time = timer.getTime
local mime = require "mime"
local json = require "lunajson"
local channel = require "channel"
local kp = require "keyPress"
local eval = require "eval"
local deckLib = require "deckLib"
local platform = system.getOS()
local ceil = math.ceil
local floor = math.floor
local auth = 0
local buff = ""
local cmdQ = {}
local cfg = ...
local server
local client = {}
local clients = 0
local clientIndex = 0
local rxTimeout = 0
local backup = {}
local restore = {}
local restoreData = {}
local updateData = {}
local chunkSize = 40000
local canvasSX, canvasSY = 0.5, 0.5
local vars = {}
local _process
require "soundManager"
require "fileLib"


local function evaluate(s)
	-- variable substitution
	for vSub in string.gmatch(s, "%$[%w_]+") do
		local v = vSub:sub(2)
		if vars[v] then
			s = s:gsub("%$"..v, vars[v])
		end
	end
	-- eval expressions in square brackets
	local _i, _j = string.find(s, "%[[^]]-%]")
	while _i do
		local sub = eval.math(s:sub(_i + 1, _j - 1))
		s = s:sub(1, _i - 1)..sub..s:sub(_j + 1)
		_i, _j = string.find(s, "%[[^]]-%]")
	end
	return s
end


-- used by pcall
local function jsonDecode(data)
	return json.decode(data)
end


local function debug(data)
	channel.main:push(json.encode({ id = "debug", data = data }))
end


local function sendAll(data)
	for k, v in pairs(client) do
		if v and v.auth and v.sock and v.sock.send then v.sock:send(data) end
	end
end


local function pushClients()
	local tcpClient = {}
	for k, v in pairs(client) do
		if v.auth then tcpClient[k] = v.user end
	end
	channel.http:push(json.encode({ id = "tcpclients", data = tcpClient }))
end


local function close(id)
	local c, s = client[id], nil
	if c then s = c.sock end
	if s and s.close then
		if c.auth then auth = auth - 1 end
		s:close()
		client[id] = nil
		clients = clients - 1
		channel.main:push(json.encode({ id = "connection", data = auth }))
		if cfg.debug then debug("Disconnected "..id) end
		pushClients()
	end
end


local function closeAll()
	for k, v in pairs(client) do
		if v and v.sock and v.sock.close then v.sock:close() end
	end
	client = {}
	clientIndex = 0
	clients = 0
	auth = 0
	pushClients()
end


local function reset()
	if server and server.close then server:close() end
	server = nil
	closeAll()
	backup = {}
	restore = {}
	restoreData = {}
	fs.remove("app2obs")
	fs.remove("obs2app")
	fs.remove("obs2dsk")
	channel.main:push(json.encode({ id = "connection", data = auth }))
end


local function backupList()
	local backupFiles = {}
	local files = fs.getDirectoryItems("Backup")
	for _, file in ipairs(files) do
		table.insert(backupFiles, file)
	end
	table.sort( backupFiles, function(a, b) return a:lower() < b:lower() end )
	local backupList
	for _, file in ipairs(backupFiles) do
		if not backupList then
			backupList = file
		else
			backupList = backupList.."\t"..file
		end
	end
	return backupList or ""
end


-- optionally receive client reference
local function process(cData, c)
	local connected = c and c.sock and c.sock.send
	local vParams = {}
	local iParams = {}
	for field in string.gmatch(cData, "[^\t]+") do
		local l = field:len()
		local i = field:find("=", 1, true)
		if i and i > 1 and i < l then
			vParams[field:sub(1, i - 1)] = field:sub(i + 1)
		else
			iParams[#iParams + 1] = field
		end
	end

	local cmd = (iParams[1] or ""):lower()

	if cmd == "play" or cmd == "sfx" then
		audio.play("Sounds/"..iParams[2])
	elseif cmd == "sfxstop" then
		audio.stop("Sounds/"..iParams[2], vParams.clean)
	elseif cmd == "key" and kp.active and vParams.app and vParams.data then
		local dir = fs.getSaveDirectory()
		if dir:sub(-1) ~= "/" then dir = dir.."/" end
		if platform == "Windows" then
			local f = floor(socket.gettime() * 1000)..".bat"
			fs.write("Win/"..f, '"'..dir..'kp.vbs" "'..vParams.app..'" "'..vParams.data..'"\r\ndel %0 *>nul')
		elseif platform == "OS X" then
			local mod, key, code
			local data = vParams.data
			local i = data:find("+", 1, true)
			if i then
				mod = data:sub(1, i-1)
				key = data:sub(i+1):upper()
			else
				mod = ""
				key = data:upper()
			end
			code = kp.codes[key]
			if code then
				local mods = { ["s"] = "shift", ["c"] = "control", ["o"] = "option", ["C"] = "command" }
				local using
				for k, v in pairs(mods) do
					if mod:find(k, 1, true) then
						if using then using = using..", " else using = "" end
						using = using..v.." down"
					end
				end
				if using then
					using = " using {"..using.."}"
				else
					using = ""
				end
				os.execute("osascript -e 'tell application \""..vParams.app.."\" to activate'")
				os.execute("osascript -e 'tell application \"System Events\" to key code "..code..using.."'")
			end
		elseif platform == "Linux" then
			-- take 2 for Linux
			os.execute("xdotool search --name \""..vParams.app.."\" windowactivate && xdotool key --delay=250 "..vParams.data)
		end
	elseif cmd == "backup" and connected then
		local cIndex = c.cIndex
		if not backup[cIndex] then backup[cIndex] = {} end
		if vParams.phase == "list" then
			c.sock:send( "backups\t"..backupList().."\n" )
		elseif vParams.phase == "begin" then
			if not fs.getInfo("Backup") then
				fs.createDirectory("Backup")
				if not fs.getInfo("Backup") then
					c.sock:send( "alert\tBackup\tError: could not create backup folder on desktop!\n" )
					return false
				end
			end
			if not vParams.name then
				c.sock:send( "alert\tBackup\tError: backup name is required to create desktop backup!\n" )
				return false
			end
			backup[cIndex].name = vParams.name
			backup[cIndex].data = ""
			if fs.getInfo("Backup/"..vParams.name) then
				c.sock:send( "backup\texists\n" )
			else
				c.sock:send( "backup\tbegin\n" )
			end
		elseif vParams.phase == "ready" then
			c.sock:send( "backup\tready\n" )
		elseif vParams.phase == "data" and vParams.chunk then
			local size = tonumber(vParams.chunk)
			local chunk, err
			if size and size > 0 then
				c.sock:settimeout(15)
				chunk, err = c.sock:receive(size)
				if chunk then
					backup[cIndex].data = (backup[cIndex].data or "")..chunk
					c.sock:send( "backup\tready\n" )
				end
				c.sock:settimeout(0)
			else
				err = "Backup : no chunk size received"
			end
			if err then
				if cfg.debug then debug("Backup Error: "..err) end
				c.sock:send( "backup\terror\t"..err.."\n" )
			end
		elseif vParams.phase == "end" then
			local jsonData = mime.unb64(backup[cIndex].data)
			if jsonData then
				local check, data = pcall(jsonDecode, jsonData)
				if check and type(data) == "table" and type(data.buttonData) == "table" and type(data.groupData) == "table" and type(data.sliderData) == "table" then
					if fs.write("Backup/"..backup[cIndex].name, jsonData) then
						c.sock:send( "alert\tBackup\tBackup was created successfully!\n" )
						c.sock:send( "backups\t"..backupList().."\n" )
					else
						c.sock:send( "alert\tBackup\tError: could not write backup data to file!\n" )
					end
				else
					c.sock:send( "alert\tBackup\tError: backup data not valid!\n" )
				end
			end
		elseif vParams.phase == "delete" and vParams.name then
			if fs.remove("Backup/"..vParams.name) then
				c.sock:send( "backups\t"..backupList().."\n" )
			else
				c.sock:send( "alert\tBackup\tAn error occurred - the backup was NOT deleted!\n" )
			end
		end
	elseif cmd == "restore" and connected then
		local cIndex = c.cIndex
		if not restore[cIndex] then restore[cIndex] = {} end
		if vParams.name then
			restoreData[cIndex] = fs.read("Backup/"..vParams.name)
			if not restoreData[cIndex] then
				c.sock:send( "alert\tRestore\tAn error occurred reading the backup file to restore!\n" )
			else
				restoreData[cIndex]:gsub("\r\n", "\n")
				restoreData[cIndex] = mime.b64(restoreData[cIndex])
				restore[cIndex].chunks = ceil( restoreData[cIndex]:len() / chunkSize )
				if cfg.debug then debug("chunks:", restore[cIndex].chunks) end
				restore[cIndex].index = 0
				c.sock:send( "restore\tbegin\n" )
			end
		elseif vParams.ready then
			if restoreData[cIndex] and restoreData[cIndex] ~= "" then
				restore[cIndex].index = restore[cIndex].index + 1
				local percent = ceil( (restore[cIndex].index / restore[cIndex].chunks) * 100 )
				local chunk = restoreData[cIndex]:sub(1, chunkSize)
				restoreData[cIndex] = restoreData[cIndex]:sub(chunkSize + 1)
				if cfg.debug then debug( "Restore "..percent.."%" ) end
				c.sock:send( "restore\tdata\t"..percent.."\t"..chunk:len().."\n"..chunk )
			else
				c.sock:send( "restore\tend\n" )
			end
		elseif vParams.error then
			restoreData[cIndex] = nil
		end
	elseif cmd == "btnup" and connected then
		local cIndex = c.cIndex
		if vParams.ready then
			if updateData[cIndex] and (updateData[cIndex].data or "") ~= "" then
				updateData[cIndex].index = updateData[cIndex].index + 1
				local percent = ceil( (updateData[cIndex].index / updateData[cIndex].chunks) * 100 )
				local chunk = updateData[cIndex].data:sub(1, chunkSize)
				updateData[cIndex].data = updateData[cIndex].data:sub(chunkSize + 1)
				if cfg.debug then debug( "Button Update "..percent.."%" ) end
				c.sock:send( "btnup\tdata\t"..percent.."\t"..chunk:len().."\n"..chunk )
			else
				c.sock:send( "btnup\tend\n" )
			end
		elseif vParams.error then
			updateData[cIndex] = nil
		end
	elseif cmd == "schedule" then
		if vParams.start then
			channel.http:push(json.encode({ id = "schedule", start = tostring(vParams.start or "") }))
		elseif iParams[2] == "stop" then
			channel.http:push(json.encode({ id = "schedule", stop = true }))
		end
	elseif cmd == "openurl" and vParams.url then
		if vParams.bg then
			local dir = fs.getSaveDirectory()
			if dir:sub(-1) ~= "/" then dir = dir.."/" end
			if platform == "Windows" then
				local f = floor(socket.gettime() * 1000)..".bat"
				fs.write("Win/"..f, '"'..dir..'bgurl.vbs" "'..vParams.url..'"\r\ndel %0 *>nul')
			else
				os.execute('curl "'..vParams.url..'" -o /dev/null')
			end
		else
			love.system.openURL(vParams.url)
		end
	elseif cmd == "launch" and vParams.app then
		if platform == "Windows" then
			local dir = fs.getSaveDirectory()
			if dir:sub(-1) ~= "/" then dir = dir.."/" end
			local f = floor(socket.gettime() * 1000)..".bat"
			fs.write("Win/"..f, 'start "" /B "'..vParams.app..'" >nul 2>&1\r\ndel %0 *>nul')
		elseif platform == "OS X" then
			os.execute('open -a "'..vParams.app..'"')
		else
			os.execute(vParams.app)
		end
	elseif cmd == "kill" and vParams.app then
		if platform == "Windows" then
			local dir = fs.getSaveDirectory()
			if dir:sub(-1) ~= "/" then dir = dir.."/" end
			local f = floor(socket.gettime() * 1000)..".bat"
			fs.write("Win/"..f, 'taskkill /f /im "'..vParams.app..'" >nul 2>&1\r\ndel %0 *>nul')
		else
			os.execute('killall "'..vParams.app..'"')
		end
	elseif cmd == "setvar" then
		for k, v in pairs(vParams) do
			vars[k] = evaluate(v)
		end
	elseif cmd == "button" then
		local xdeck = vParams.deck
		if not xdeck then xdeck = vars.__livedeck__ end
		local button = vParams.id
		if xdeck and button then
			local _cmd, msg = deckLib.press(xdeck, button)
			if not _cmd then
				if cfg.debug then debug(msg) end
			elseif _cmd ~= "" then
				_process(_cmd)
			end
		end
	elseif cmd == "XXX_hotkey" and connected and c.cIndex == 1 then -- no longer used
		-- save command data for hotkey triggering
		local hkIndex = tonumber(vParams.index) or 0
		local size = tonumber(vParams.chunk)
		if hkIndex > 0 and size and size > 0 then
			c.sock:settimeout(15)
			local chunk, err = c.sock:receive(size)
			if chunk then
				fs.write("Actions/"..hkIndex..".txt", mime.unb64(chunk))
			end
			c.sock:settimeout(0)
		end
		hkIndex = hkIndex + 1
		if hkIndex < 101 then
			c.sock:send("hotkey\t"..hkIndex.."\n")
		end
	elseif cmd == "user" and iParams[2] then
		if iParams[2] ~= "" then
			c.user = iParams[2]
			pushClients()
		end
	elseif cmd == "draw" then
		local update
		local canvasData = {}
		if vParams.w then
			local w = tonumber(vParams.w) or 0
			if w < 640 then w = 640 end
			canvasSX = 640 / w
			debug(string.format("Stream width = %d (%0.1f)", w, canvasSX))
		end
		if vParams.h then
			local h = tonumber(vParams.h) or 0
			if h < 360 then h = 360 end
			canvasSY = 360 / h
			debug(string.format("Stream height = %d (%0.1f)", h, canvasSY))
		end
		if vParams.clear then
			canvasData.clear = true
			update = true
		elseif vParams.undo then
			canvasData.undo = true
			update = true
		end
		if vParams.r and vParams.g and vParams.b then
			canvasData.r = vParams.r or 0
			canvasData.g = vParams.g or 0.8
			canvasData.b = vParams.b or 0
			update = true
		end
		if vParams.s then
			canvasData.size = tonumber(vParams.s) or 5
			update = true
		end
		local x = tonumber(vParams.x)
		local y = tonumber(vParams.y)
		if x and y then
			canvasData.x = x * canvasSX
			canvasData.y = y * canvasSY
			update = true
		end
		if vParams.stop then
			canvasData.stop = true
			update = true
		end
		if update then channel.main:push(json.encode({ id = "canvas", data = canvasData })) end
	end
end


_process = function(msg, c)
	for cData in string.gmatch(msg, "[^\n]+") do
		cData = evaluate(cData)
		if cfg.debug then debug(cData) end
		local lData = cData:lower()
		if not lData:find("\tcondition=0", 1, true) then
			if (cfg.sfx and (lData:sub(1, 5) == "play\t" or lData:sub(1, 4) == "sfx\t" or lData:sub(1, 8) == "sfxstop\t"))
			or lData:sub(1, 5) == "draw\t"
			or lData:sub(1, 4) == "key\t"
			or lData:sub(1, 7) == "backup\t"
			or lData:sub(1, 8) == "restore\t"
			or lData:sub(1, 7) == "hotkey\t" 
			or lData:sub(1, 5) == "user\t"
			or lData:sub(1, 6) == "btnup\t"
			or lData:sub(1, 9) == "schedule\t"
			or lData:sub(1, 8) == "openurl\t"
			or lData:sub(1, 7) == "launch\t"
			or lData:sub(1, 5) == "kill\t"
			or lData:sub(1, 7) == "setvar\t"
			or lData:sub(1, 7) == "button\t" then
				-- process locally
				local delay = cData:match("\tdelay=%d+")
				if delay then
					local targetTime = floor(time() * 1000) + delay:sub(8)
					cmdQ[#cmdQ + 1] = { time = targetTime, cData = cData }
				else
					process(cData, c)
				end
			elseif cData:len() > 1 then
				buff = buff..cData.."\n"
			end
		end
	end
end


local function listen()
	if not server then
		server = socket.bind("*", cfg.port, 0)
		if server then
			server:settimeout(0)
			server:setoption("reuseaddr", true)
			if cfg.debug then debug(string.format("UPDeck listening on port %d", cfg.port)) end
		end
		return nil
	end

	
	-- check for new connection
	local c = server:accept()
	if c then
		local _ip, _port = c:getpeername()
		local id = _ip..":".._port
		clientIndex = clientIndex + 1
		client[id] = { sock = c, cIndex = clientIndex, auth = false, user = _ip }
		clients = clients + 1
		if cfg.debug then debug("Connected "..id) end
		c:settimeout(0)
		c:setoption("tcp-nodelay", true)
		c:setoption("keepalive", true)
		if cfg.tout > 0 then rxTimeout = os.time() + cfg.tout end
		if clients == 1 then
			fs.remove("app2obs")
			fs.remove("obs2app")
			fs.remove("obs2dsk")
			restoreData[clientIndex] = nil
		end
	end
	

	-- check for obs data for app
	if fs.getInfo("obs2app") then
		local data = fs.read("obs2app")
		if data then
			if fs.remove("obs2app") then sendAll(data) end
		end
	end

	-- check for app data for obs
	local msg, stop

	for id, c in pairs(client) do
		msg = nil
		stop = false
		repeat
			local line, err = c.sock:receive("*l")
			if not line then
				if err ~= "timeout" then
					close(id)
				end
			else		
				if line:sub(1, 1) == "*" then
					line = line:sub(2)
					stop = true
				end
				-- handle any special cases
				if line:sub(1, 7) == "button\t" then
					-- ignore
				else
					if not msg then msg = "" else msg = msg.."\n" end
					msg = msg..line
				end
			end
		until not line or stop

		if msg then
			if cfg.debug then debug(string.format("Received data from %s", id)) end

			-- check if timeout is active
			if cfg.tout > 0 then rxTimeout = os.time() + cfg.tout end

			if not c.auth then
				local _ip, _port = c.sock:getpeername()
				-- at least 1 user has problems using the ip and port auth check..
				-- for now, drop back to simpler passwword check
				msg = mime.unb64(msg) or ""
				local i = msg:find("-", 1, true)
				local j = msg:reverse():find("-", 1, true)
				if i and j and msg:sub(i+1, -j-1) == cfg.pass then
					c.auth = true
					c.sock:send( ":)\n" )
					auth = auth + 1
					channel.main:push(json.encode({ id = "connection", data = auth }))
					pushClients()
					if clients == 1 then
						-- clear folder - ready to import actions from current client
						fs.rDelete("Actions")
						fs.createDirectory("Actions")
						buff = "obsfx\t"..(cfg.sfx and 0 or 1).."\n"
					end
				else
					if cfg.debug then debug(string.format("Bad password   IP: %s   Port: %s", _ip or "?", _port or "?")) end
					c.sock:send( ":(\n" )
					channel.main:push(json.encode({ id = "connection", data = auth }))
				end
			else
				-- inject obs data for desktop : hotkeys
				-- delay parameter will have already elapsed, so disable
				if fs.getInfo("obs2dsk") then
					local data = fs.read("obs2dsk")
					if data and fs.remove("obs2dsk") then msg = msg.."\n"..data:gsub("delay=", "xdelay=") end
				end
				_process(msg, c)
			end
		end -- msg
	end -- each client
	if buff and buff ~= "" and not fs.getInfo("app2obs") then
		if fs.write("app2obs", buff) then buff = "" end
	end
end


while true do
	love.audio.update()

	-- process delayed commands
	local t = floor(time() * 1000)
	for i = #cmdQ, 1, -1 do
		if t >= cmdQ[i].time then
			process(cmdQ[i].cData)
			table.remove(cmdQ, i)
		end
	end

	local chData
	while true do
		chData = channel.tcp:pop()
		if not chData then break end
		chData = json.decode(chData) or {}
		if not chData.id then break end
		if chData.id == "buff" and chData.data then
			buff = buff..chData.data
		elseif chData.id == "schedule" and chData.data then
			_process(chData.data, nil)
		elseif chData.id == "xdeck" and chData.data then
			if chData.deck then vars.__livedeck__ = chData.deck end
			_process(chData.data, nil)
		elseif chData.id == "cfg" and chData.data then
			cfg = chData.data
		elseif chData.id == "reset" then
			reset()
		elseif chData.id == "btnupdate" and chData.data then
			local update = chData.data
			if update.client and client[update.client] and update.data then
				local sock = client[update.client].sock
				if sock and sock.send then
					local cIndex = client[update.client].cIndex
					updateData[cIndex] = { index = 0, data = mime.b64(json.encode(update.data)):gsub("\n", "") }
					updateData[cIndex].chunks = ceil( updateData[cIndex].data:len() / chunkSize )
					sock:send("btnup\tinit\n")
				end
			end
		elseif chData.id == "volupdate" then
			local update = chData.data
			if update.client and client[update.client] and update.data then
				local sock = client[update.client].sock
				if sock and sock.send then
					sock:send("volup\t"..mime.b64(json.encode(update.data, "___null___")):gsub("\n", "").."\n")
				end
			end
		end
	end

	listen()

	-- check for timeout
	if cfg.tout > 0 and clients > 0 and os.time() > rxTimeout then
		if cfg.debug then debug("Timeout : resetting connection") end
		reset()
	end

	timer.sleep(0.02)
end